"""
Some common bitwise manipulations.
"""


def set_bit(target, bit):
    """
    Returns target but with the given bit set to 1
    """
    return target | (1 << bit)

def clear_bit(target, bit):
    """
    Returns target but with the given bit set to 0
    """
    return target & ~(1 << bit)

def bit_high(value, bit):
    """
    Returns whether the bit specified is set high in value
    e.g. bit_high(64, 6) == True    (64 = 0b01000000, so bit 6 is high)
         bit_high(64, 2) == False
    """
    return value & (1 << bit) > 0

def bit_low(value, bit):
    """
    Returns whether the bit specified is set low in value
    e.g. bit_high(64, 6) == False    (64 = 0b01000000, so bit 6 is high)
         bit_high(64, 2) == True
    """
    return value & (1 << bit) == 0


########### Byte operations ############################################################################################

def lower_nibble(value):
    """
    Returns the value of the lower nibble (4 bits) of a byte (i.e. a value in range 0-15)
    """
    return value & 0x0F

def upper_nibble(value):
    """
    Returns the value of the upper nibble (4 bits) of a byte (i.e. a value in range 0-15)
    """
    return (value & 0xF0) >> 4


########### Word operations ############################################################################################

def set_high_byte(target, hi_byte):
    """
    Sets the high (second least significant) byte of a 16 bit (or longer) target value (*not* a little endian word!) to
    hi_byte
    """
    return (target & 0x00FF) + (hi_byte << 8)

def set_low_byte(target, lo_byte):
    """
    Sets the low byte of a 16 bit (or longer) target value (*not* a little endian word!) to lo_byte
    """
    return (target & 0xFF00) + lo_byte



"""
Meta code that generates the run_instr function for the MOS6502 from the instruction tables.  This replaced a possibly
nicer pythonic way of doing the instruction decoding that turned out to be slow and not very compatible with Cython
optimization.

Run this file and copy its output to the end of mos6502.pyx
"""

from .instructions import INSTRUCTION_SET, AddressModes


def generate_cpu_instruction_table(tab_level = 1, spaces_per_tab=4):
    first_code = True
    indent = " " * tab_level * spaces_per_tab
    print(indent + "cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):")
    tab_level += 1
    indent = " " * tab_level * spaces_per_tab
    print(indent + "# *********************************************************************** #")
    print(indent + "# **          Autogenerated by meta.py - Do not edit directly          ** #")
    print(indent + "# *********************************************************************** #")
    print(indent)
    print(indent + "cdef int cycles=0, arg, immediate=False, m")

    for _, instr_set in INSTRUCTION_SET.items():
        for mode, instr in instr_set.modes.items():
            if type(instr.bytecode) == list:
                opcodes = instr.bytecode
            else:
                opcodes = [instr.bytecode]

            print(indent + "{}if opcode==".format("" if first_code else "el"), end="")
            for i, opcode in enumerate(opcodes):
                print("0x{:02X}".format(opcode), end="")
                if i < len(opcodes) - 1:
                    print(" or opcode==", end="")
                else:
                    print(":")
            if first_code:
                first_code = False
            tab_level += 1
            indent = " " * tab_level * spaces_per_tab
            print(indent + "cycles = {}".format(int(instr.cycles)))
            print(generate_addressing(indent, mode, instr.cycles))
            print(indent + "cycles += self._{}(arg, immediate)".format(instr_set.name))
            tab_level -= 1
            indent = " " * tab_level * spaces_per_tab
    print(indent)
    print(indent + "return cycles")


def generate_addressing(indent, mode, cycles):
    cde = indent
    if mode == AddressModes.IMPLIED:
        # all implied instructions are self-contained 1-byte instructions
        cde += "arg = ARG_NONE"
    elif mode == AddressModes.ACCUMULATOR:
        # the instructions with accumulator targets usually have a memory address as a target, if
        # None is supplied, they target the accumulator
        cde += "arg = ARG_NONE"
    elif mode == AddressModes.RELATIVE:
        # all relative instructions are control flow instructions that take a relative address
        print(indent + "arg = data[0]")
    elif mode == AddressModes.IMMEDIATE:
        cde += "arg = data[0]\n"
        cde += indent + "immediate = True"
    else:
        # memory loading modes:
        #cde += "address = ARG_NONE\n" + indent
        # arg is an address
        if mode == AddressModes.ZEROPAGE:
            cde += "arg = data[0]"
        elif mode == AddressModes.ZEROPAGE_X:
            cde += "arg = (data[0] + self.X) & 0xFF"
        elif mode == AddressModes.ZEROPAGE_Y:
            cde += "arg = (data[0] + self.Y) & 0xFF"
        elif mode == AddressModes.ABSOLUTE:
            cde += "arg = self._from_le(data)"
        elif mode == AddressModes.ABSOLUTE_X:
            cde += "arg = (self._from_le(data) + self.X) & 0xFFFF"
            if cycles > int(cycles):
                cde += "\n" + indent + "cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0"
        elif mode == AddressModes.ABSOLUTE_Y:
            cde += "arg = (self._from_le(data) + self.Y) & 0xFFFF"
            if cycles > int(cycles):
                cde += "\n" + indent + "cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0"
        elif mode == AddressModes.INDIRECT:
            # only used for jmp instructions
            # has the jump indirect bug [12] which means it cannot cross page boundaries and instead
            # wraps around, e.g. read from 0x12ff reads 0x12ff and 0x1200
            # address =
            cde += "arg = self._read_word(self._from_le(data), wrap_at_page=True)"
            # address = self._from_le(self.memory.read_block(self._from_le(data), bytes=2))
        elif mode == AddressModes.INDIRECT_X:
            cde += "arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)"
        elif mode == AddressModes.INDIRECT_Y:
            cde += "m = self._read_word(data[0], wrap_at_page=True)\n"
            cde += indent + "arg = (m + self.Y) & 0xFFFF"
            if cycles > int(cycles):
                cde += "\n" + indent + "cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0"
    return cde


generate_cpu_instruction_table(tab_level=1, spaces_per_tab=4)


import logging

#from nes import LOG_MEMORY


OAM_SIZE_BYTES = 256


class MemoryBase:
    """
    Basic memory controller interface
    """
    def __init__(self):
        pass

    def read(self, address):
        raise NotImplementedError()

    def read_block(self, address, bytes):
        rval = bytearray(bytes)
        for i in range(bytes):
            rval[i] = self.read(address + i)
        return rval

    def write(self, address, value):
        raise NotImplementedError()

    def print(self, address, bytes, width_bytes=16):
        for off in range(bytes):
            if off % width_bytes == 0:
                if off > 0:
                    print()
                print("${0:04X}:   ".format(address + off), end="")
            v = self.read(address + off)
            print("{0:02X}".format(v), end=" ")


class BigEmptyRAM(MemoryBase):
    """
    Just a big empty bank of 64kB of RAM
    """
    def __init__(self):
        super().__init__()
        self.ram = bytearray(2 ** 16)

    def read(self, address):
        return self.ram[address]

    def read_block(self, address, bytes):
        # this is so simple, we can do a more efficient read_block
        return self.ram[address:(address + bytes)]

    def write(self, address, value):
        self.ram[address] = value


class NESMappedRAM(MemoryBase):
    """
    NES memory following NES CPU memory map pattern

    References:
        [1] CPU memory map:  https://wiki.nesdev.com/w/index.php/CPU_memory_map
    """
    RAM_SIZE = 0x800            # 2kB of internal RAM
    NUM_PPU_REGISTERS = 8       # number of ppu registers

    RAM_END = 0x0800            # NES main ram to here
    PPU_END = 0x4000            # PPU registers to here
    APU_END = 0x4018            # APU registers (+OAM DMA reg) to here
    APU_UNUSED_END = 0x4020     # generally unused APU and I/O functionality
    OAM_DMA = 0x4014            # OAM DMA register address
    CONTROLLER1 = 0x4016        # port for controller (read controller 1 / write both controllers)
    CONTROLLER2 = 0x4017        # port for controller 2 (read only, writes to this port go to the APU)
    CART_START = 0x4020         # start of cartridge address space

    def __init__(self, ppu=None, apu=None, cart=None, controller1=None, controller2=None, interrupt_listener=None):
        super().__init__()
        self.ram = bytearray(self.RAM_SIZE)  # 2kb of internal RAM
        self.ppu = ppu
        self.apu = apu
        self.cart = cart
        self.controller1 = controller1
        self.controller2 = controller2
        self.interrupt_listener = interrupt_listener

        # internal variable used for open bus behaviour
        self._last_bus = 0

    def read(self, address):
        """
        Read one byte of memory from the NES address space
        """
        if address < self.RAM_END:    # RAM and its mirrors
            region = "ram"
            value = self.ram[address % self.RAM_SIZE]
        elif address < self.PPU_END:  # PPU registers
            region = "ppu"
            register_ix = address % self.NUM_PPU_REGISTERS
            if self.ppu is not None:
                value = self.ppu.read_register(register_ix)
            else:
                value = 0
        elif address < self.APU_END:
            region = "apu/oam"
            if address == self.OAM_DMA and self.ppu:
                # write only
                value = 0
            elif address == self.CONTROLLER1:
                value = (self.controller1.read_bit() & 0b00011111) + (0x40 & 0b11100000)
                #print("{:08b}".format(value))
                #print("{:08b}".format(self._last_bus))
                # todo: deal with open bus behaviour of upper control lines
            elif address == self.CONTROLLER2:
                # todo: deal with open bus behaviour of upper control lines
                value = (self.controller2.read_bit() & 0b00011111) + (0x40 & 0b11100000)
            else:
                # todo: APU registers
                value = 0
        elif address < self.APU_UNUSED_END:
            # todo: generally unused APU and I/O functionality
            region = "apu-unused"
            value = 0
        else:
            # cartridge space; pass this to the cart, which might do its own mapping
            region = "cart"
            value = self.cart.read(address)

        #logging.log(LOG_MEMORY, "read {:04X}  (= {:02X})  region={:10s}".format(address, value, region), extra={"source": "mem"})
        return value

    def write(self, address, value):
        """
        Write one byte of memory in the NES address space
        """
        #logging.log(LOG_MEMORY, "write {:02X} --> {:04X}".format(value, address), extra={"source": "mem"})

        if address < self.RAM_END:    # RAM and its mirrors
            self.ram[address % self.RAM_SIZE] = value
        elif address < self.PPU_END:  # PPU registers
            register_ix = address % self.NUM_PPU_REGISTERS
            if self.ppu:
                self.ppu.write_register(register_ix, value)
        elif address < self.APU_END:
            if address == self.OAM_DMA:
                if self.ppu:
                    self.run_oam_dma(value)
            elif address == self.CONTROLLER1:
                self.controller1.set_strobe(value)
                self.controller2.set_strobe(value)
            else:
                # todo: APU registers
                pass
        elif address < self.APU_UNUSED_END:
            # todo: generally unused APU and I/O functionality
            pass
        else:
            # cartridge space; pass this to the cart, which might do its own mapping
            self.cart.write(address, value)

    def run_oam_dma(self, page):
        """
        OAM DMA copies an entire page (wrapping at the page boundary if the start address in ppu's oam_addr is not zero)
        from RAM to ppu OAM.  This also causes the cpu to pause for 513 or 514 cycles.
        :param page:
        :return:
        """
        logging.debug("OAM DMA from page {:02X}".format(page), extra={"source": "mem"})
        # done in two parts to correctly account for wrapping at page end
        data_block = bytearray(256)
        data_block[self.ppu.oam_addr:OAM_SIZE_BYTES] = self.read_block(page << 8, OAM_SIZE_BYTES - self.ppu.oam_addr)
        data_block[0:self.ppu.oam_addr] = self.read_block(page << 8, self.ppu.oam_addr)
        self.ppu.write_oam(data_block)

        #self.ppu.oam[self.ppu.oam_addr:OAM_SIZE_BYTES] = self.read_block(page << 8, OAM_SIZE_BYTES - self.ppu.oam_addr)
        #self.ppu.oam[0:self.ppu.oam_addr] = self.read_block(page << 8, self.ppu.oam_addr)
        # tell the interrupt listener that the CPU should pause due to OAM DMA
        self.interrupt_listener.raise_oam_dma_pause()


class NESVRAM(MemoryBase):
    """
    NES video (PPU) RAM, following the PPU memory map pattern

    References:
        [1] PPU memory map: https://wiki.nesdev.com/w/index.php/PPU_memory_map
    """
    PATTERN_TABLE_SIZE_BYTES = 4096   # provided by the rom
    NAMETABLES_SIZE_BYTES = 2048
    PALETTE_SIZE_BYTES = 32
    NAMETABLE_LENGTH_BYTES = 1024  # single nametime is this big

    # memory map
    NAMETABLE_START = 0x2000
    ATTRIBUTE_TABLE_OFFSET = 0x3C0  # offset of the attribute table from the start of the corresponding nametable
    PALETTE_START = 0x3F00

    # Mirror patterns
    # The mirror pattern specifies the underlying nametable at locations 0x2000, 0x2400, 0x2800 and 0x3200
    MIRROR_HORIZONTAL = (0, 0, 1, 1)
    MIRROR_VERTICAL = (0, 1, 0, 1)
    #MIRROR_SINGLE = (0, 0, 0, 0)
    MIRROR_FOUR_SCREEN = (0, 1, 2, 3)
    #MIRROR_DIAGONAL = (0, 1, 1, 0)
    #MIRROR_L_SHAPED = (0, 1, 1, 1)
    #MIRROR_3_SCREEN_V = (0, 2, 1, 2)
    #MIRROR_3_SCREEN_H = (0, 1, 2, 2)
    #MIRROR_3_SCREEN_DIAG = (0, 1, 1, 2)

    def __init__(self, cart, nametable_size_bytes=2048):
        super().__init__()
        self.cart = cart
        # self._pattern_table = bytearray(self.PATTERN_TABLE_SIZE_BYTES)
        self._nametables = bytearray(nametable_size_bytes)
        self.palette_ram = bytearray(self.PALETTE_SIZE_BYTES)
        self.nametable_mirror_pattern = cart.nametable_mirror_pattern

    def decode_address(self, address):
        if address < self.NAMETABLE_START:
            # pattern table - provided by the rom
            #return self._pattern_table, address
            return self.cart.chr_mem, address % len(self.cart.chr_mem)  # todo: need something better here via the read_ppu/wrtie_ppu in order to implement mappers
        elif address < self.PALETTE_START:
            # nametable
            page = int((address - self.NAMETABLE_START) / self.NAMETABLE_LENGTH_BYTES)  # which nametable?
            offset = (address - self.NAMETABLE_START) % self.NAMETABLE_LENGTH_BYTES  # offset in that table

            # some of the pages (e.g. 2 and 3) are mirrored, so for these, find the underlying
            # namepage that they point to based on the mirror pattern
            true_page = self.nametable_mirror_pattern[page]
            return self._nametables, true_page * self.NAMETABLE_LENGTH_BYTES + offset
        else:
            # palette table
            if address == 0x3F10 or address == 0x3F14 or address == 0x3F18 or address == 0x3F1C:
                # "addresses $3F10/$3F14/$3F18/$3F1C are mirrors of $3F00/$3F04/$3F08/$3F0C"
                # (https://wiki.nesdev.com/w/index.php/PPU_palettes)
                address -= 0x10
            return self.palette_ram, address % self.PALETTE_SIZE_BYTES

    def read(self, address):
        memory, address_decoded = self.decode_address(address)
        value = memory[address_decoded]
        #logging.log(LOG_MEMORY, "read {:04X} [{:04X}]  (= {:02X})".format(address, address_decoded, value), extra={"source": "vram"})
        return value

    def write(self, address, value):
        #logging.log(LOG_MEMORY, "write {:02X} --> {:04X}".format(value, address), extra={"source": "vram"})
        memory, address = self.decode_address(address)
        memory[address] = value



#import pyximport; pyximport.install()

from .system cimport OAM_DMA
from .ppu cimport OAM_SIZE_BYTES, NUM_PPU_REGISTERS


cdef class MemoryBase:
    """
    Basic memory controller interface
    """
    def __init__(self):
        pass

    cdef unsigned char read(self, int address):
        raise NotImplementedError()

    cdef void write(self, int address, unsigned char value):
        raise NotImplementedError()


###### Main Memory #####################################################################################################
cdef class NESMappedRAM(MemoryBase):
    """
    NES memory following NES CPU memory map pattern

    References:
        [1] CPU memory map:  https://wiki.nesdev.com/w/index.php/CPU_memory_map
    """
    def __init__(self, ppu=None, apu=None, cart=None, controller1=None, controller2=None, interrupt_listener=None):
        super().__init__()
        self.ppu = ppu
        self.apu = apu
        self.cart = cart
        self.controller1 = controller1
        self.controller2 = controller2
        self.interrupt_listener = interrupt_listener

        # internal variable used for open bus behaviour
        self._last_bus = 0

    cdef unsigned char read(self, int address):
        """
        Read one byte of memory from the NES address space
        """
        cdef unsigned char value

        if address < RAM_END:    # RAM and its mirrors
            value = self.ram[address % RAM_SIZE]
        elif address < PPU_END:  # PPU registers
            value = self.ppu.read_register(address % NUM_PPU_REGISTERS)
        elif address < APU_END:
            if address == OAM_DMA_REG:
                # write only
                value = 0
            elif address == CONTROLLER1:
                value = (self.controller1.read_bit() & 0b00011111) + (0x40 & 0b11100000)
                # todo: deal with open bus behaviour of upper control lines
            elif address == CONTROLLER2:
                # todo: deal with open bus behaviour of upper control lines
                value = (self.controller2.read_bit() & 0b00011111) + (0x40 & 0b11100000)
            else:
                # this is a read of the apu, pass it on directly
                value = self.apu.read_register(address)
        elif address < APU_UNUSED_END:
            # todo: generally unused APU and I/O functionality
            value = 0
        else:
            # cartridge space; pass this to the cart, which might do its own mapping
            value = self.cart.read(address)

        #logging.log(LOG_MEMORY, "read {:04X}  (= {:02X})  region={:10s}".format(address, value, region), extra={"source": "mem"})
        return value

    cdef void write(self, int address, unsigned char value):
        """
        Write one byte of memory in the NES address space
        """
        #logging.log(LOG_MEMORY, "write {:02X} --> {:04X}".format(value, address), extra={"source": "mem"})
        cdef int register_ix

        if address < RAM_END:    # RAM and its mirrors
            self.ram[address % RAM_SIZE] = value
        elif address < PPU_END:  # PPU registers
            register_ix = address % NUM_PPU_REGISTERS
            self.ppu.write_register(register_ix, value)
        elif address < APU_END:
            if address == OAM_DMA_REG:
                self.run_oam_dma(value)
            elif address == CONTROLLER1:
                self.controller1.set_strobe(value)
                self.controller2.set_strobe(value)
            else:
                # this is a write of an apu register; pass it on to the apu
                #print("apu cycles at write below:", self.apu.cycles)
                #print("apu write: {:02X} -> {:04X}".format(value, address))
                self.apu.write_register(address, value)
        elif address < APU_UNUSED_END:
            # todo: generally unused APU and I/O functionality
            pass
        else:
            # cartridge space; pass this to the cart, which might do its own mapping
            self.cart.write(address, value)

    cdef void run_oam_dma(self, int page):
        """
        OAM DMA copies an entire page (wrapping at the page boundary if the start address in ppu's oam_addr is not zero)
        from RAM to ppu OAM.  This also causes the cpu to pause for 513 or 514 cycles.
        :param page:
        :return:
        """
        # done in two parts to correctly account for wrapping at page end
        cdef unsigned char data_block[OAM_SIZE_BYTES]
        cdef int i, addr_base, oam_addr

        oam_addr = self.ppu.oam_addr
        addr_base = page << 8
        for i in range(OAM_SIZE_BYTES):
            data_block[(i + oam_addr) & 0xFF] = self.read( addr_base + i )

        # pass with the size here to avoid zero-terminating as if it were a string
        self.ppu.write_oam(data_block[:OAM_SIZE_BYTES])
        # tell the interrupt listener that the CPU should pause due to OAM DMA
        self.interrupt_listener.raise_dma_pause(OAM_DMA)

    def __getstate__(self):
        # annoyingly, Cython pickles char arrays as null terminated strings, so we have to do this manually
        ram = self.ram[:RAM_SIZE]
        state = (ram, self._last_bus, self.apu, self.ppu, self.controller1, self.controller2, self.interrupt_listener)
        return state

    def __setstate__(self, state):
        (ram, _last_bus, apu, ppu, controller1, controller2, interrupt_listener) = state
        for i in range(RAM_SIZE):
            self.ram[i] = ram[i]
        self._last_bus = _last_bus
        self.apu = apu
        self.ppu = ppu
        self.controller1 = controller1
        self.controller2 = controller2
        self.interrupt_listener = interrupt_listener
        return state



###### VRAM ############################################################################################################

cdef class NESVRAM(MemoryBase):
    """
    NES video (PPU) RAM, following the PPU memory map pattern

    References:
        [1] PPU memory map: https://wiki.nesdev.com/w/index.php/PPU_memory_map
    """
    def __init__(self, cart):
        super().__init__()
        self.cart = cart

    cdef unsigned char read(self, int address):
        cdef unsigned char value
        cdef int page, offset, true_page

        if address < NAMETABLE_START:
            # pattern table - provided by the rom
            value = self.cart.read_ppu(address)
        elif address < PALETTE_START:
            # nametable
            page = int((address - NAMETABLE_START) / NAMETABLE_LENGTH_BYTES)    # which nametable?
            offset = (address - NAMETABLE_START) % NAMETABLE_LENGTH_BYTES  # offset in that table

            # some of the pages (e.g. 2 and 3) are mirrored, so for these, find the underlying
            # namepage that they point to based on the mirror pattern
            true_page = self.cart.nametable_mirror_pattern[page]
            value = self._nametables[true_page * NAMETABLE_LENGTH_BYTES + offset]
        else:
            # palette table
            if address == 0x3F10 or address == 0x3F14 or address == 0x3F18 or address == 0x3F1C:
                # "addresses $3F10/$3F14/$3F18/$3F1C are mirrors of $3F00/$3F04/$3F08/$3F0C"
                # (https://wiki.nesdev.com/w/index.php/PPU_palettes)
                address -= 0x10
            return self.palette_ram[address % PALETTE_SIZE_BYTES]
        return value

    cdef void write(self, int address, unsigned char value):
        cdef int page, offset, true_page

        if address < NAMETABLE_START:
            # pattern table - provided by the rom
            self.cart.write_ppu(address, value)
        elif address < PALETTE_START:
            # nametable
            page = int((address - NAMETABLE_START) / NAMETABLE_LENGTH_BYTES)  # which nametable?
            offset = (address - NAMETABLE_START) % NAMETABLE_LENGTH_BYTES  # offset in that table

            # some of the pages (e.g. 2 and 3) are mirrored, so for these, find the underlying
            # namepage that they point to based on the mirror pattern
            true_page = self.cart.nametable_mirror_pattern[page]
            self._nametables[true_page * NAMETABLE_LENGTH_BYTES + offset] = value
        else:
            # palette table
            if address == 0x3F10 or address == 0x3F14 or address == 0x3F18 or address == 0x3F1C:
                # "addresses $3F10/$3F14/$3F18/$3F1C are mirrors of $3F00/$3F04/$3F08/$3F0C"
                # (https://wiki.nesdev.com/w/index.php/PPU_palettes)
                address -= 0x10
            self.palette_ram[address % PALETTE_SIZE_BYTES] = value



#import pyximport; pyximport.install()

from .system cimport OAM_DMA, DMC_DMA, DMC_DMA_DURING_OAM_DMA
from nes.instructions import INSTRUCTION_SET, NamedInstruction, AddressModes

cdef class MOS6502:
    """
    Software emulator for MOS Technologies 6502 CPU.  The chip in the NES is a modified version of the 6502 that does
    not support BCD mode (which makes it a bit simpler).

    References:
       [1] https://www.masswerk.at/6502/6502_instruction_set.html
       [2] http://www.obelisk.me.uk/6502/addressing.html
       [3] https://www.csh.rit.edu/~moffitt/docs/6502.html

       [4] V-flag:  http://www.6502.org/tutorials/vflag.html
       [5] Decimal mode: http://www.6502.org/tutorials/decimal_mode.html
       [6] PC: http://6502.org/tutorials/6502opcodes.html#PC
       [7] http://forum.6502.org/viewtopic.php?t=1708

       [8] Assembler online: https://www.masswerk.at/6502/assembler.html
       [9] CPU startup sequence:  https://www.pagetable.com/?p=410
       [10] B-flag: http://wiki.nesdev.com/w/index.php/Status_flags#The_B_flag
       [11] V-flag: http://www.righto.com/2012/12/the-6502-overflow-flag-explained.html
       [12] JMP indirect 'bug':  http://forum.6502.org/viewtopic.php?f=2&t=6170

       [13] Undocumented opcodes: http://nesdev.com/undocumented_opcodes.txt
       [14] Undocumented opcodes: http://www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes

       [15] Amazing timing data: http://www.6502.org/users/andre/petindex/local/64doc.txt

       [16] Undocumented opcodes: http://hitmen.c02.at/files/docs/c64/NoMoreSecrets-NMOS6510UnintendedOpcodes-20162412.pdf

    """
    def __init__(self, memory, undocumented_support_level=1, aax_sets_flags=False, stack_underflow_causes_exception=True):

        # memory is user-supplied object with read and write methods, allowing for memory mappers, bank switching, etc.
        self.memory = memory

        # create a map from bytecodes to functions for the instruction set
        self.instructions = self._make_bytecode_dict()
        self.make_instruction_data_tables()

        # User-accessible registers
        self.A = 0  # accumulator
        self.X = 0  # X register
        self.Y = 0  # Y register

        # Indirectly accessible registers
        self.PC = 0     # program counter (not valid until a reset() has occurred)
        self.SP = 0     # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)

        # Flags
        # (stored as an 8-bit register in the CPU)
        # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
        self.N = 0
        self.V = 0
        self.D = 0
        self.I = 0
        self.Z = 0
        self.C = 0

        # CPU cycles since the processor was started or reset
        self.cycles_since_reset = 0

        # control behaviour of the cpu
        self.aax_sets_flags = aax_sets_flags
        self.undocumented_support_level = undocumented_support_level
        self.stack_underflow_causes_exception = stack_underflow_causes_exception

        # reset the cpu
        self.reset()


    def reset(self):
        """
        Resets the CPU
        """
        # read the program counter from the RESET_VECTOR_ADDR
        self.PC = self._read_word(RESET_VECTOR_ADDR, wrap_at_page=False)

        # clear the registers
        self.A = 0  # accumulator
        self.X = 0  # X register
        self.Y = 0  # Y register

        # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)
        # reset does three stack pops, so this is set to 0xFD after reset [9]
        self.SP = 0xFD

        # Flags
        # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
        self.N = False
        self.V = False
        self.D = False
        self.I = True   # CPU should "come up with interrupt disable bit set" [7]
        self.Z = False
        self.C = False

        # CPU cycles since the processor was started or reset
        # reset takes 7 cycles [9]
        self.cycles_since_reset = 7

    cdef int dma_pause(self, int pause_type, int count):
        cdef int cycles
        if pause_type == OAM_DMA:
            cycles = OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2
        elif pause_type == DMC_DMA:
            #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
            cycles = 4
        elif pause_type == DMC_DMA_DURING_OAM_DMA:
            #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
            cycles = 2 * count
        self.cycles_since_reset += cycles
        return cycles

    @property
    def cycles_since_reset(self):
        return self.cycles_since_reset

    def _make_bytecode_dict(self):
        """
        Translates the instruction sets into a bytecode->instr dictionary and links the instruction opcodes
        to the functions in this object that execute them (this is done by instruction name).
        """
        instructions = [None] * 256
        for _, instr_set in INSTRUCTION_SET.items():
            for mode, instr in instr_set.modes.items():
                if type(instr.bytecode) == list:
                    # some of the undocumented opcodes have multiple aliases, so in that case a list of
                    # opcodes is supplied
                    for bytecode in instr.bytecode:
                        instructions[bytecode] = NamedInstruction(name=instr_set.name,
                                                                  bytecode=bytecode,
                                                                  mode=mode,
                                                                  size_bytes=instr.size_bytes,
                                                                  cycles=instr.cycles,
                                                                 )
                else:
                    instructions[instr.bytecode] = NamedInstruction(name=instr_set.name,
                                                                    bytecode=instr.bytecode,
                                                                    mode=mode,
                                                                    size_bytes=instr.size_bytes,
                                                                    cycles=instr.cycles,
                                                                   )
        return instructions

    cdef void make_instruction_data_tables(self):
        """
        Fills in some data tables about the instructions in cdef variables; replaces the more pythonic version
        make_bytecode_dict.
        """
        for _, instr_set in INSTRUCTION_SET.items():
            for mode, instr in instr_set.modes.items():
                bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
                for bytecode in bytecodes:
                    self.instr_size_bytes[bytecode] = instr.size_bytes

    def format_instruction(self, instr, data, caps=True):
        """
        Formats an instruction for logline in the form of nestest logs
        """
        line = ""
        name = instr.name if not caps else instr.name.upper()
        line += '{} '.format(name)
        if instr.mode == AddressModes.IMMEDIATE:
            line += '#${0:02X}'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE:
            line += '${0:02X}'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE_X:
            line += '${0:02X},X'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE_Y:
            line += '${0:02X},Y'.format(data[0])
        elif instr.mode == AddressModes.ABSOLUTE:
            line += '${0:04X}'.format(self._from_le(data))
        elif instr.mode == AddressModes.ABSOLUTE_X:
            line += '${0:04X},X'.format(self._from_le(data))
        elif instr.mode == AddressModes.ABSOLUTE_Y:
            line += '${0:04X},Y'.format(self._from_le(data))
        elif instr.mode == AddressModes.INDIRECT_X:
            line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
        elif instr.mode == AddressModes.INDIRECT_Y:
            line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
        elif instr.mode == AddressModes.INDIRECT:
            line += '(${0:04X})'.format(self._from_le(data))
        elif instr.mode == AddressModes.IMPLIED:
            line += ''
        elif instr.mode == AddressModes.ACCUMULATOR:
            line += 'A'
        elif instr.mode == AddressModes.RELATIVE:
            line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
        return line

    def log_line(self, ppu_line="---", ppu_pixel="---"):
        """
        Generates a log line in the format of nestest
        """
        str = "{0:04X}  ".format(self.PC)
        bytecode = self.memory.read(self.PC)

        instr = self.instructions[bytecode]
        data = bytearray(2)
        for i in range(instr.size_bytes - 1):
            data[i] = self.memory.read(self.PC + 1 + i)

        str += "{0:02X} ".format(bytecode)
        str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "
        str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "
        str += self.format_instruction(instr, data)

        while len(str) < 48:
            str += " "

        str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,
                                                                                         self.X,
                                                                                         self.Y,
                                                                                         self._status_to_byte(b_flag=0),
                                                                                         self.SP,
                                                                                         ppu_line,
                                                                                         ppu_pixel,
                                                                                         self.cycles_since_reset
                                                                                        )
        return str

    def set_reset_vector(self, reset_vector):
        """
        Sets the reset vector (at fixed mem address), which tells the program counter where to start on a reset
        :param reset_vector: 16-bit address to which PC will be set after a reset
        """
        self.memory.write(RESET_VECTOR_ADDR, reset_vector & 0x00FF)  # low byte
        self.memory.write(RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte

    cdef int _from_le(self, unsigned char* data):
        """
        Create an integer from a little-endian two byte array
        :param data: two-byte little endian array
        :return: an integer in the range 0-65535
        """
        return (data[HI_BYTE] << 8) + data[LO_BYTE]

    cdef int _read_word(self, int addr, int wrap_at_page):
        """
        Read a word at an address and return it as an integer
        If wrap_at_page is True then if the address is 0xppFF, then the word will be read from 0xppFF and 0xpp00
        """
        cdef unsigned char byte_lo, byte_hi

        if wrap_at_page and (addr & 0xFF) == 0xFF:
            # will wrap at page boundary
            byte_lo = self.memory.read(addr)
            byte_hi = self.memory.read(addr & 0xFF00)  # read the second (hi) byte from the start of the page
        else:
            #data = self.memory.read_block(addr, bytes=2)
            byte_lo = self.memory.read(addr)
            byte_hi = self.memory.read(addr + 1)
        return byte_lo + (byte_hi << 8)

    cdef int trigger_nmi(self):
        """
        Trigger a non maskable interrupt (NMI)
        """
        self._interrupt(NMI_VECTOR_ADDR, is_brk=False)
        self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
        return INTERRUPT_REQUEST_CYCLES

    cdef int trigger_irq(self):
        """
        Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
        """
        if not self.I:
            self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)
            self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
            return INTERRUPT_REQUEST_CYCLES
        else:
            return 0  # ignored!

    cdef int run_next_instr(self):
        """
        Decode and run the next instruction at the program counter, updating the number of processor
        cycles that have elapsed.  Instructions are taken as atomic, since the 6502 will complete them
        before it responds to interrupts.
        """
        cdef int i, cycles
        cdef unsigned char bytecode
        cdef unsigned char data[2]

        bytecode = self.memory.read(self.PC)
        # surprisingly, loop unrolling here helped quite a bit
        if self.instr_size_bytes[bytecode] >= 2:
            data[0] = self.memory.read(self.PC + 1)
        if self.instr_size_bytes[bytecode] == 3:
            data[1] = self.memory.read(self.PC + 2)

        # upon retrieving the opcode, the 6502 immediately increments PC by the opcode size:
        self.PC += self.instr_size_bytes[bytecode]

        cycles = self.run_instr(bytecode, data)
        self.cycles_since_reset += cycles
        return cycles

    cdef unsigned char _status_to_byte(self, bint b_flag):
        """
        Puts the status register into an 8-bit value.  Bit 6 is set high always, bit 5 (the "B flag") is set according
        to the b_flag argument.  This should be high if called from an instruction (PHP or BRK) instruction, low if
        the call came from a hardware interrupt (IRQ or NMI).
        """
        return (  self.N * SR_N_MASK
                + self.V * SR_V_MASK
                + SR_X_MASK              # the unused bit should always be set high
                + b_flag * SR_B_MASK     # the "B flag" should be set high if called from PHP or BRK [10] low o/w
                + self.D * SR_D_MASK
                + self.I * SR_I_MASK
                + self.Z * SR_Z_MASK
                + self.C * SR_C_MASK) & 0xFF   # this final and forces it to be treated as unsigned

    cdef void _status_from_byte(self, unsigned char sr_byte):
        """
        Sets the processor status from an 8-bit value as found on the stack.
        Bit 5 (B flag) is NEVER set in the status register (but is on the stack), so it is ignored here
        """
        self.N = (sr_byte & SR_N_MASK) > 0
        self.V = (sr_byte & SR_V_MASK) > 0
        self.D = (sr_byte & SR_D_MASK) > 0
        self.I = (sr_byte & SR_I_MASK) > 0
        self.Z = (sr_byte & SR_Z_MASK) > 0
        self.C = (sr_byte & SR_C_MASK) > 0

    cdef void push_stack(self, unsigned char v):
        """
        Push a byte value onto the stack
        """
        self.memory.write(STACK_PAGE + self.SP, v)
        self.SP = (self.SP - 1) & 0xFF

    cdef unsigned char pop_stack(self):
        """
        Pop (aka 'pull' in 6502 parlance) a byte from the stack
        """
        cdef unsigned char v
        if self.SP == 0xFF and self.stack_underflow_causes_exception:
            raise OverflowError("Stack underflow")
        self.SP = (self.SP + 1) & 0xFF
        v = self.memory.read(STACK_PAGE + self.SP)
        return v

    cdef int _neg(self, int v):
        """
        Is the value of v negative in 2s complement (i.e. is bit 7 high)?
        :return: true if v is negative (true if bit 7 of v is 1)
        """
        return (v & 0b10000000) > 0

    cdef int _from_2sc(self, unsigned char v):
        """
        Convert a 2's complement number to a signed integer
        """
        #neg = (v & 0b10000000) > 0
        #return (v & 0b01111111) if not neg else (v & 0b01111111) - 128
        return (v & 0b01111111) - (v & 0b10000000)

    cdef void _set_zn(self, int v):
        """
        Sets the Z and N flags from a result value v
        """
        self.Z = (v & 0xFF) == 0  # only care about the bottom 8 bits being zero
        self.N = (v & 0b10000000) > 0

    ##################  The instructions (and their helper functions)  #################################################

    cdef int _adc(self, int arg, int immediate):
        """
        Add to accumulator with carry.  Result goes into accumulator.
        :param arg: value to be added to accumulator (cannot be None)
        """
        cdef unsigned char v
        cdef int result

        v = self.memory.read(arg) if not immediate else arg

        # binary mode (standard)
        result = self.A + v + self.C
        self.C = result > 255
        self.V = (self._neg(self.A) == self._neg(v)) and (self._neg(self.A) != self._neg(result))

        # status
        # setting N and Z flags in the usual way is valid in BCD mode because "the N flag contains the high bit of
        # the result of the instruction" [5] (having said that, BCD mode is removed on the NES)
        self._set_zn(result)

        # result
        self.A = result & 0xFF

        return 0

    cdef int _and(self, int arg, int immediate):
        """
        Bitwise AND of accumulator and arg.  Result goes into accumulator.
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self.A = self.A & v
        self._set_zn(self.A)
        return 0

    cdef int _asl(self, int addr, int _):
        """
        Shift left one bit.
        """
        cdef unsigned char v, res_8bit
        cdef int result
        v = self.A if addr==ARG_NONE else self.memory.read(addr)
        result = v << 1

        # status
        self.C = result > 255    # top bit of original is put into carry flag
        self._set_zn(result)

        res_8bit = result & 0xFF

        if addr==ARG_NONE:
            # target the accumulator
            self.A = res_8bit
        else:
            self.memory.write(addr, res_8bit)
        return 0

    cdef int _jump_relative(self, int condition, int offset_2sc):
        """
        Jump by offset_2sc (in 2s complement) if the condition is true
        :return: number of extra cycles incurred by this branch conditional
        """
        cdef int extra_cycles, prev_pc_page

        extra_cycles = 0
        if condition:
            #print("jump")
            extra_cycles = 1  # taking a branch takes at least 1 extra cycle
            prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
            self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
            if prev_pc_page != self.PC & 0xFF00:
                # but it takes two extra cycles if the memory page changes
                extra_cycles = 2
        #else:
        #    print("no jump")
        return extra_cycles

    cdef int _bcc(self, int offset, int _):
        """
        Branch on carry clear
        """
        return self._jump_relative(not self.C, offset)

    cdef int _bcs(self, int offset, int _):
        """
        Branch on carry set
        """
        return self._jump_relative(self.C, offset)

    cdef int _beq(self, int offset, int _):
        """
        Branch if zero flag is set
        """
        return self._jump_relative(self.Z, offset)

    cdef int _bit(self, int addr, int _):
        """
        Manipulates the status register by setting the N and V flags to those bits of the value in the address
        given, and sets the zero flag if A & v == 0
        """
        cdef unsigned char v
        v = self.memory.read(addr)
        self.N = (v & SR_N_MASK) > 0
        self.V = (v & SR_V_MASK) > 0
        self.Z = (self.A & v) == 0
        return 0

    cdef int _bmi(self, int offset, int _):
        """
        Branch on result minus (i.e. negative flag N is set)
        """
        return self._jump_relative(self.N, offset)

    cdef int _bne(self, int offset, int _):
        """
        Branch if zero flag is not set
        """
        return self._jump_relative(not self.Z, offset)

    cdef int _bpl(self, int offset, int _):
        """
        Branch on result positive (i.e. negative flag N is not set)
        """
        #print(self.N, not self.N)
        return self._jump_relative(not self.N, offset)

    cdef _interrupt(self, int interrupt_vector, int is_brk):
        """
        Interrupt routine, followed (with variations) by NMI, IRQ and BRK
          1)
        """
        cdef int v, addr
        cdef unsigned char sr
        # push PC (+1 if BRK) to the stack, high bit first
        v = self.PC + (1 if is_brk else 0)
        self.push_stack((v & 0xFF00) >> 8)  # high byte
        self.push_stack(v & 0xFF)  # low byte
        # push the processor status to the stack
        # BUT note that the B flag ON THE STACK COPY ONLY is now set
        sr = self._status_to_byte(b_flag=True if is_brk else False)
        self.push_stack(sr)
        addr = self._read_word(interrupt_vector, wrap_at_page=False)
        self.PC = addr
        # Set the interrupt disable flag.  But there seems to be disagreement in the sources as to whether or not the
        # I flag is set if this is a BRK instruction.
        # The following sources have it set:
        #   https://www.masswerk.at/6502/6502_instruction_set.html#BRK
        #   blargg's instruction tests (15-brk)
        # The following have it unaffected:
        #   http://www.obelisk.me.uk/6502/reference.html#BRK
        #   http://www.6502.org/tutorials/6502opcodes.html#BRK
        #   https://www.csh.rit.edu/~moffitt/docs/6502.html#BRK
        # Since blargg's tests run on a hardware NES, we follow that behaviour
        self.I = True

    cdef int _brk(self, int _, int __):
        """
        Force break, which simulates an interrupt request (IRQ).
        BRK, unlike other interrupts (IRQ and NMI), pushes PC + 1 to the stack (high byte first as usual).
        The reason for this may have been to allow brk to be dropped in in place of two byte instructions to allow
        debugging, but it is a quirk of the BRK instruction.  BRK also sets the B flag in the value of SR pushed to
        the stack
        """
        self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=True)
        return 0

    cdef int _bvc(self, int offset, int _):
        """
        Branch on on overflow clear (V == 0)
        """
        return self._jump_relative(not self.V, offset)

    cdef int _bvs(self, int offset, int _):
        """
        Branch on on overflow set (V == 1)
        """
        return self._jump_relative(self.V, offset)

    cdef int _clc(self, int _, int __):
        """
        Clear carry flag (set C:=0)
        """
        self.C = False
        return 0

    cdef int _cld(self, int _, int __):
        """
        Clear decimal flag (set D:=0)
        """
        self.D = False
        return 0

    cdef int _cli(self, int _, int __):
        """
        Clear interrupt disable flag (set I:=0)
        """
        self.I = False
        return 0

    cdef int _clv(self, int _, int __):
        """
        Clear value flag (set V:=0)
        """
        self.V = False
        return 0

    cdef void _compare(self, int v0, int v):
        self._set_zn(v0 - v)
        self.C = v0 >= v

    cdef int _cmp(self, int arg, int immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.A, v)
        return 0

    cdef int _cpx(self, int arg, int immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.X, v)
        return 0

    cdef int _cpy(self, int arg, int immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.Y, v)
        return 0

    cdef int _dec(self, int addr, int _):
        """
        Decrement memory in addr by 1
        :param addr:
        :param _:
        :return:
        """
        cdef unsigned char v
        v = (self.memory.read(addr) - 1) & 0xFF
        self._set_zn(v)
        self.memory.write(addr, v)
        return 0

    cdef int _dex(self, int _, int __):
        """
        Decrement X by 1
        :param addr:
        :param _:
        :return:
        """
        self.X = (self.X - 1) & 0xFF
        self._set_zn(self.X)
        return 0

    cdef int _dey(self, int _, int __):
        """
        Decrement Y by 1
        :param addr:
        :param _:
        :return:
        """
        self.Y = (self.Y - 1) & 0xFF
        self._set_zn(self.Y)
        return 0

    cdef int _eor(self, int arg, int immediate):
        """
        XOR A with value and put result back into A
        :param addr:
        :param _:
        :return:
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self.A = v ^ self.A
        self._set_zn(self.A)
        return 0

    cdef int _inc(self, int addr, int _):
        """
        Increment memory in addr by 1
        :param addr:
        :param _:
        :return:
        """
        cdef unsigned char v
        v = (self.memory.read(addr) + 1) & 0xFF
        self._set_zn(v)
        self.memory.write(addr, v)
        return 0

    cdef int _inx(self, int _, int __):
        """
        Increment X by 1
        :param addr:
        :param _:
        :return:
        """
        self.X = (self.X + 1) & 0xFF
        self._set_zn(self.X)
        return 0

    cdef int _iny(self, int _, int __):
        """
        Increment Y by 1
        :param addr:
        :param _:
        :return:
        """
        self.Y = (self.Y + 1) & 0xFF
        self._set_zn(self.Y)
        return 0

    cdef int _jmp(self, int addr, int _):
        """
        Jump to the (16 bit) address addr.  In the case of jump instructions jmp and jsr,
        "absolute" addressing is more like immediate addressing, in that the jump goes directly
        to the memory address specified rather than reading the value there and using that (which
        is "indirect" addressing in the case of jumps, so this function behaves always like it got
        an immediate address
        """
        self.PC = addr
        return 0

    cdef int _jsr(self, int addr, int _):
        """
        Jump to a subroutine at the value in memory[addr].  Addr here is an absolute (16 bit) location.
        In the case of jump instructions jmp and jsr,
        "absolute" addressing is more like immediate addressing, in that the jump goes directly
        to the memory address specified rather than reading the value there and using that (which
        is "indirect" addressing in the case of jumps, so this function behaves always like it got
        an immediate address
        """
        # save PC - 1 to the stack, high byte first
        cdef int v
        v = self.PC - 1
        self.push_stack((v & 0xFF00) >> 8)
        self.push_stack(v & 0x00FF)
        self.PC = addr
        return 0

    cdef int _lda(self, int arg, int immediate):
        """
        Load A from value / memory
        :return:
        """
        self.A = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.A)
        return 0

    cdef int _ldx(self, int arg, int immediate):
        """
        Load X from value / memory
        :return:
        """
        self.X = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.X)
        return 0

    cdef int _ldy(self, int arg, int immediate):
        """
        Load Y from value / memory
        :return:
        """
        self.Y = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.Y)
        return 0

    cdef int _lsr(self, int addr, int _):
        """
        Logical shift right
        :param addr:
        :param _:
        :return:
        """
        cdef unsigned char v, result
        v = self.A if addr==ARG_NONE else self.memory.read(addr)
        result = (v >> 1) & 0xFF

        # status
        self.C = v & 0x01
        self.N = False
        self.Z = result == 0

        if addr==ARG_NONE:
            # target the accumulator
            self.A = result
        else:
            self.memory.write(addr, result)
        return 0

    cdef int _nop(self, int _, int __):
        """
        No-op.  This one is easy.
        :return:
        """
        return 0

    cdef int _ora(self, int arg, int immediate):
        """
        Bitwise OR with accumulator; result put into accumulator.
        :return:
        """
        cdef unsigned char v
        v = self.memory.read(arg) if not immediate else arg
        self.A = self.A | v
        self._set_zn(self.A)
        return 0

    cdef int _pha(self, int _, int __):
        """
        Push A onto stack
        :param _:
        :param __:
        :return:
        """
        self.push_stack(self.A)
        return 0

    cdef int _php(self, int _, int __):
        """
        Push status register onto stack
        :param _:
        :param __:
        :return:
        """
        cdef unsigned char v

        v = self._status_to_byte(b_flag=True)  # b-flag should be set True in this case [10]
        self.push_stack(v)
        return 0

    cdef int _pla(self, int _, int __):
        """
        Pull A from stack
        :param _:
        :param __:
        :return:
        """
        self.A = self.pop_stack()
        self._set_zn(self.A)
        return 0

    cdef int _plp(self, int _, int __):
        """
        Pull processor status from the stack
        :return:
        """
        cdef unsigned char v
        v = self.pop_stack()
        self._status_from_byte(v)
        return 0

    cdef int _rol(self, int addr, int _):
        """
        Rotate one bit left (including carry bit in the rotation, i.e. C rotates into lsb and msb rotates into C)
        :return:
        """
        cdef unsigned char v
        cdef int result

        v = self.A if addr==ARG_NONE else self.memory.read(addr)
        result = (v << 1) + self.C

        # status
        self.C = result > 255
        self._set_zn(result)

        if addr==ARG_NONE:
            # target the accumulator
            self.A = result & 0xFF
        else:
            self.memory.write(addr, result & 0xFF)
        return 0

    cdef int _ror(self, int addr, int _):
        """
        Rotate one bit right (including carry bit in the rotation, i.e. C rotates into msb and lsb rotates into C)
        :return:
        """
        cdef unsigned char v
        cdef int result
        v = self.A if addr==ARG_NONE else self.memory.read(addr)
        result = (v >> 1) + (self.C << 7)

        # status
        self.C = v & 0x01
        self._set_zn(result)

        if addr==ARG_NONE:
            # target the accumulator
            self.A = result & 0xFF
        else:
            self.memory.write(addr, result & 0xFF)
        return 0

    cdef int _rti(self, int _, int __):
        """
        Return from an interrupt.  Gets the PC of the last execution point and the flag register from the
        stack.  (PC is stored on the stack, not PC-1 as with jsr/rts)
        :return:
        """
        cdef unsigned char flags, addr_lo, addr_hi
        flags = self.pop_stack()
        addr_lo = self.pop_stack()
        addr_hi = self.pop_stack()
        self._status_from_byte(flags)
        self.PC = (addr_hi << 8) + addr_lo
        return 0

    cdef int _rts(self, int _, int __):
        """
        Returns from a subroutine (see jsr for the jump subroutine)
        :return:
        """
        cdef unsigned char addr_lo, addr_hi
        # fetch PC-1 from the stack, low byte first
        addr_lo = self.pop_stack()
        addr_hi = self.pop_stack()
        # restore PC to the stack
        self.PC = (addr_hi << 8) + addr_lo + 1
        return 0

    cdef int _sbc(self, int arg, int immediate):
        """
        Subtract memory plus carry flag from A and put the result in A (i.e. A := A - M - ~C).
        The carry flag is set if A - M - ~C >= 0, i.e. if no borrow is necessary (allows chaining of
        subtract instructions).
        The overflow (V) flag is set if there is a (signed) overflow.
        Overflow occurs when the result is either too negative or too positive to be held in the available 8 bits.
        This can happen in two cases:
          1) when A is +ve and v is -ve; this should create a +ve number, so if result is -ve that is overflow
          2) when A is -ve and v is +ve; this should create a -ve number, so if result is +ve that is overflow
        """
        cdef unsigned char v, borrow
        cdef int result
        v = self.memory.read(arg) if not immediate else arg
        borrow = not self.C   # incoming borrow is indicated by not Carry  (borrow is either 0 or 1)

        result = self.A - v - borrow     # arithmetic is done unsigned
        self.C = result >= 0  # not Carry == borrow; borrow if res<0, so C==not(res<0) == res>=0
        # overflow - see notes in description above
        # I think this is correct and that the borrow doesn't matter here
        self.V = (self._neg(self.A) != self._neg(v)) and (self._neg(v) == self._neg(result))
        self._set_zn(result)
        self.A = result & 0xFF
        return 0

    cdef int _sec(self, int _, int __):
        """
        Set carry flag
        """
        self.C = True
        return 0

    cdef int _sed(self, int _, int __):
        """
        Set decimal flag
        """
        self.D = True
        return 0

    cdef int _sei(self, int _, int __):
        """
        Set interrupt disable flag
        """
        self.I = True
        return 0

    cdef int _sta(self, int addr, int _):
        """
        Store A in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.A)
        return 0

    cdef int _stx(self, int addr, int _):
        """
        Store X in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.X)
        return 0

    cdef int _sty(self, int addr, int _):
        """
        Store Y in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.Y)
        return 0

    cdef int _tax(self, int _, int __):
        """
        Transfer A to X
        :return:
        """
        self.X = self.A
        self._set_zn(self.X)
        return 0

    cdef int _tay(self, int _, int __):
        """
        Transfer A to Y
        :return:
        """
        self.Y = self.A
        self._set_zn(self.Y)
        return 0

    cdef int _tsx(self, int _, int __):
        """
        Transfer SP to X
        :return:
        """
        self.X = self.SP
        self._set_zn(self.X)
        return 0

    cdef int _txa(self, int _, int __):
        """
        Transfer X to A
        :return:
        """
        self.A = self.X
        self._set_zn(self.A)
        return 0

    cdef int _txs(self, int _, int __):
        """
        Transfer X to SP
        :return:
        """
        self.SP = self.X
        return 0

    cdef int _tya(self, int _, int __):
        """
        Transfer Y to A
        :return:
        """
        self.A = self.Y
        self._set_zn(self.A)
        return 0

    ################################## The undocumented instructions (see [13] and others) #############################

    def _dop(self, _, __):
        """
        Undocumented.
        Double NOP instruction.  Luckily it is easy to implement!
        """
        return 0

    def _top(self, _, __):
        """
        Undocumented.
        Triple NOP instruction.
        """
        return 0

    def _kil(self, _, __):
        """
        Undocumented.
        Shuts down the processor.  No recovery is possible after this has executed, short of a reset.
        """
        if self.undocumented_support_level >= 1:
            raise ValueError("KIL instruction.  Processor halted.")
        return 0

    def _lax(self, arg, immediate):
        """
        Undocumented.
        Load A and X from value / memory
        :return:
        """
        if self.undocumented_support_level >= 1:
            v = self.memory.read(arg) if not immediate else arg
            self.A = v
            self.X = v
            self._set_zn(self.A)
        return 0

    def _aax(self, addr, _):
        """
        Undocumented.
        AND X register with accumulator and store result in memory.
        Status flags: N,Z  (or are they?  nestest does not think so...)
        """
        if self.undocumented_support_level >= 1:
            v = self.A & self.X
            self.memory.write(addr, v)
            if self.aax_sets_flags:
                # the behaviour of this seems unclear, nestest does not think it does this, [13] does
                self._set_zn(v)
        return 0

    def _dcp(self, addr, _):
        """
        Undocumented.
        This opcode DECs the contents of a memory location and then CMPs the result with the A register.
        Equivalent to DEC oper, CMP oper [14]
        """
        if self.undocumented_support_level >= 1:
            v = (self.memory.read(addr) - 1) & 0xFF
            self.memory.write(addr, v)
            self._compare(self.A, v)
        return 0

    def _isc(self, addr, _):
        """
        Undocumented.
        Increase memory by one, then subtract memory from accumulator (with borrow). Status flags: N,V,Z,C
        """
        if self.undocumented_support_level >= 1:
            v = (self.memory.read(addr) + 1) & 0xFF
            self.memory.write(addr, v)
            self._sbc(v, immediate=True)
        return 0

    def _slo(self, addr, _):
        """
        Undocumented.
        Shift left one bit in memory, then OR accumulator with memory.  Status flags: N,Z,C
        Equivalent to ASL, ORA
        """
        if self.undocumented_support_level >= 1:
            self._asl(addr, ARG_NONE)
            self._ora(self.memory.read(addr), immediate=True)
        return 0

    def _rla(self, addr, _):
        """
        Undocumented.
        RLA ROLs the contents of a memory location and then ANDs the result with the accumulator.
        Equivalent to ROL, AND
        """
        if self.undocumented_support_level >= 1:
            self._rol(addr, ARG_NONE)
            self._and(self.memory.read(addr), immediate=True)
        return 0

    def _rra(self, addr, _):
        """
        Undocumented.
        RRA RORs the contents of a memory location and then ADCs the result with the accumulator.
        Equivalent to ROR, ADC
        """
        if self.undocumented_support_level >= 1:
            self._ror(addr, ARG_NONE)
            self._adc(self.memory.read(addr), immediate=True)
        return 0

    def _sre(self, addr, _):
        """
        Undocumented.
        Shift right one bit in memory, then EOR accumulator with memory.
        Equivalent to LSR, EOR
        """
        if self.undocumented_support_level >= 1:
            self._lsr(addr, ARG_NONE)
            self._eor(self.memory.read(addr), immediate=True)
        return 0

    ##################### The arcane undocumented instructions (see [13] and others) ###################################

    def _arr(self, arg, _):
        """
        Undocumented.  Mostly used for piracy (jk).
        ANDs the contents of the A register with an immediate value and then RORs the result, and check bit 5 and 6:
            If both bits are 1: set C, clear V.
            If both bits are 0: clear C and V.
            If only bit 5 is 1: set V, clear C.
            If only bit 6 is 1: set C and V.
        Equivalent to AND oper, ROR A
        """
        if self.undocumented_support_level >= 2:
            self._and(arg, immediate=True)
            self._ror(ARG_NONE, ARG_NONE)
            if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):
                self.C = True
                self.V = False
            elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:
                self.C = False
                self.V = False
            elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:
                self.C = False
                self.V = True
            elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):
                self.C = True
                self.V = True
        return 0

    def _asr(self, arg, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A register with an immediate value and then LSRs the result.
        Equivalent to AND oper, LSR A
        """
        if self.undocumented_support_level >= 2:
            self._and(arg, immediate=True)
            self._lsr(ARG_NONE, ARG_NONE)
        return 0

    def _atx(self, arg, _):
        """
        Undocumented.
        FAILS BLARGG'S INSTRUCTION TEST 03-immediate (both variants here)

        Following http://nesdev.com/undocumented_opcodes.txt:
            AND byte with accumulator, then transfer accumulator to X register.
            Status flags: N,Z

        Following http://www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes (where this is OAL):
            ORs the A register with #$EE, ANDs the result with an immediate value, and then stores the result in
            both A and X.
            Equivalent to ORA #$EE, AND oper, TAX

        The latter causes blargg's test 03-immediate to fail for opcode 0xAB
        """
        if self.undocumented_support_level >= 2:
            #self.A &= arg
            #self.X = self.A
            #self._set_zn(self.A)
            self._and(arg | 0xEE, immediate=True)
            self.X = self.A
        return 0

    def _aac(self, arg, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A register with an immediate value and then
        moves bit 7 of A into the Carry flag.  This opcode works basically
        identically to AND #immed. except that the Carry flag is set to the same
        state that the Negative flag is set to.  [14]
        """
        if self.undocumented_support_level >= 2:
            self.A = self.A & arg
            self._set_zn(self.A)
            self.C = (self.A & 0b10000000) > 0
        return 0

    def _axa(self, addr, _):
        """
        UNTESTED
        Undocumented.
        Sores the result of A AND X AND the high byte of the target address of the operand +1 in memory [14, 15]
        WARNING:  There seem to be multiple conflicting descriptions of this opcode. (see e.g. [13])
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            v = (self.X & self.A) & hp1
            self.memory.write(addr, v)
        return 0

    def _axs(self, arg, immediate):
        """
        FAILS BLARGG'S INSTRUCTION TEST 03-immediate
        Undocumented.
        AND the contents of the A and X registers (without changing the contents of either register) and
        stores the result in memory. [16]
        """
        self.A = self.memory.read(arg) if not immediate else arg

        if self.undocumented_support_level >= 2:

            if immediate:
                # AND X register with accumulator and store result in X regis-ter, then subtract byte from X register
                # (without borrow).  Status flags: N,Z,C   (http://nesdev.com/undocumented_opcodes.txt)
                self.X = self.A & self.X
                t = self.X - arg
                self.C = t >= 0
                self.X = t & 0xFF
                self._set_zn(self.X)
            else:
                self.memory.write(arg, self.A & self.X)
                self._set_zn(self.A & self.X)
        return 0

    def _lar(self, addr, _):
        """
        UNTESTED
        Undocumented.
        AND memory with stack pointer, transfer result to accumulator, X register and stack pointer. [16]
        """
        if self.undocumented_support_level >= 2:
            v = self.memory.read(addr) & self.SP
            self.A = v
            self.X = v
            self.SP = v
        return 0

    def _sxa(self, addr, _):
        """
        FAILS BLARGG'S TEST 07-abs_xy
        Undocumented.
        AND X register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            self.memory.write(addr, self.X & hp1)
        return 0

    def _sya(self, addr, _):
        """
        FAILS BLARGG'S TEST 07-abs_xy
        Undocumented.
        AND Y register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
        """
        if self.undocumented_support_level >= 2:
            v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
            self.memory.write(addr, v & 0xFF)
        return 0

    def _xaa(self, arg, _):
        """
        UNTESTED
        Undocumented.
        This opcode ORs the A register with CONST, ANDs the result with X. ANDs the result with an immediate value, and
        then stores the result in A.
        Known to be extremely unstable, including temperature dependent! [16]
        """
        if self.undocumented_support_level >= 2:
            CONST = 0x00
            self.A = (self.A | CONST) & self.X & arg
        return 0

    def _xas(self, addr, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A and X registers (without changing the contents
        of either register) and transfers the result to the stack pointer. It then ANDs that result with the
        contents of the high byte of the target address of the operand +1 and stores that final result in
        memory
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            self.SP = self.A & self.X
            self.memory.write(addr, self.SP & hp1)
        return 0

    cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):
        # *********** AUTOGENERATED BY meta.py - DO NOT EDIT DIRECTLY ***********
        cdef int cycles=0, arg, immediate=False, m

        #print("{:02X} {:02X} {:02X}".format(opcode, data[0], data[1]))

        if opcode==0x69:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._adc(arg, immediate)
        elif opcode==0x65:
            cycles = 3
            arg = data[0]
            cycles += self._adc(arg, immediate)
        elif opcode==0x75:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._adc(arg, immediate)
        elif opcode==0x6D:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._adc(arg, immediate)
        elif opcode==0x7D:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._adc(arg, immediate)
        elif opcode==0x79:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._adc(arg, immediate)
        elif opcode==0x61:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._adc(arg, immediate)
        elif opcode==0x71:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._adc(arg, immediate)
        elif opcode==0x29:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._and(arg, immediate)
        elif opcode==0x25:
            cycles = 3
            arg = data[0]
            cycles += self._and(arg, immediate)
        elif opcode==0x35:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._and(arg, immediate)
        elif opcode==0x2D:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._and(arg, immediate)
        elif opcode==0x3D:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._and(arg, immediate)
        elif opcode==0x39:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._and(arg, immediate)
        elif opcode==0x21:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._and(arg, immediate)
        elif opcode==0x31:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._and(arg, immediate)
        elif opcode==0x0A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._asl(arg, immediate)
        elif opcode==0x06:
            cycles = 5
            arg = data[0]
            cycles += self._asl(arg, immediate)
        elif opcode==0x16:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._asl(arg, immediate)
        elif opcode==0x0E:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._asl(arg, immediate)
        elif opcode==0x1E:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._asl(arg, immediate)
        elif opcode==0x90:
            cycles = 2
            arg = data[0]

            cycles += self._bcc(arg, immediate)
        elif opcode==0xB0:
            cycles = 2
            arg = data[0]

            cycles += self._bcs(arg, immediate)
        elif opcode==0xF0:
            cycles = 2
            arg = data[0]

            cycles += self._beq(arg, immediate)
        elif opcode==0x24:
            cycles = 3
            arg = data[0]
            cycles += self._bit(arg, immediate)
        elif opcode==0x2C:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._bit(arg, immediate)
        elif opcode==0x30:
            cycles = 2
            arg = data[0]

            cycles += self._bmi(arg, immediate)
        elif opcode==0xD0:
            cycles = 2
            arg = data[0]

            cycles += self._bne(arg, immediate)
        elif opcode==0x10:
            cycles = 2
            arg = data[0]

            cycles += self._bpl(arg, immediate)
        elif opcode==0x00:
            cycles = 7
            arg = ARG_NONE
            cycles += self._brk(arg, immediate)
        elif opcode==0x50:
            cycles = 2
            arg = data[0]

            cycles += self._bvc(arg, immediate)
        elif opcode==0x70:
            cycles = 2
            arg = data[0]

            cycles += self._bvs(arg, immediate)
        elif opcode==0x18:
            cycles = 2
            arg = ARG_NONE
            cycles += self._clc(arg, immediate)
        elif opcode==0xD8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._cld(arg, immediate)
        elif opcode==0x58:
            cycles = 2
            arg = ARG_NONE
            cycles += self._cli(arg, immediate)
        elif opcode==0xB8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._clv(arg, immediate)
        elif opcode==0xC9:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._cmp(arg, immediate)
        elif opcode==0xC5:
            cycles = 3
            arg = data[0]
            cycles += self._cmp(arg, immediate)
        elif opcode==0xD5:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._cmp(arg, immediate)
        elif opcode==0xCD:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._cmp(arg, immediate)
        elif opcode==0xDD:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._cmp(arg, immediate)
        elif opcode==0xD9:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._cmp(arg, immediate)
        elif opcode==0xC1:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._cmp(arg, immediate)
        elif opcode==0xD1:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._cmp(arg, immediate)
        elif opcode==0xE0:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._cpx(arg, immediate)
        elif opcode==0xE4:
            cycles = 3
            arg = data[0]
            cycles += self._cpx(arg, immediate)
        elif opcode==0xEC:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._cpx(arg, immediate)
        elif opcode==0xC0:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._cpy(arg, immediate)
        elif opcode==0xC4:
            cycles = 3
            arg = data[0]
            cycles += self._cpy(arg, immediate)
        elif opcode==0xCC:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._cpy(arg, immediate)
        elif opcode==0xC6:
            cycles = 5
            arg = data[0]
            cycles += self._dec(arg, immediate)
        elif opcode==0xD6:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._dec(arg, immediate)
        elif opcode==0xCE:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._dec(arg, immediate)
        elif opcode==0xDE:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._dec(arg, immediate)
        elif opcode==0xCA:
            cycles = 2
            arg = ARG_NONE
            cycles += self._dex(arg, immediate)
        elif opcode==0x88:
            cycles = 2
            arg = ARG_NONE
            cycles += self._dey(arg, immediate)
        elif opcode==0x49:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._eor(arg, immediate)
        elif opcode==0x45:
            cycles = 3
            arg = data[0]
            cycles += self._eor(arg, immediate)
        elif opcode==0x55:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._eor(arg, immediate)
        elif opcode==0x4D:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._eor(arg, immediate)
        elif opcode==0x5D:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._eor(arg, immediate)
        elif opcode==0x59:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._eor(arg, immediate)
        elif opcode==0x41:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._eor(arg, immediate)
        elif opcode==0x51:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._eor(arg, immediate)
        elif opcode==0xE6:
            cycles = 5
            arg = data[0]
            cycles += self._inc(arg, immediate)
        elif opcode==0xF6:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._inc(arg, immediate)
        elif opcode==0xEE:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._inc(arg, immediate)
        elif opcode==0xFE:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._inc(arg, immediate)
        elif opcode==0xE8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._inx(arg, immediate)
        elif opcode==0xC8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._iny(arg, immediate)
        elif opcode==0x4C:
            cycles = 3
            arg = self._from_le(data)
            cycles += self._jmp(arg, immediate)
        elif opcode==0x6C:
            cycles = 5
            arg = self._read_word(self._from_le(data), wrap_at_page=True)
            cycles += self._jmp(arg, immediate)
        elif opcode==0x20:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._jsr(arg, immediate)
        elif opcode==0xA9:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._lda(arg, immediate)
        elif opcode==0xA5:
            cycles = 3
            arg = data[0]
            cycles += self._lda(arg, immediate)
        elif opcode==0xB5:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._lda(arg, immediate)
        elif opcode==0xAD:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._lda(arg, immediate)
        elif opcode==0xBD:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._lda(arg, immediate)
        elif opcode==0xB9:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._lda(arg, immediate)
        elif opcode==0xA1:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._lda(arg, immediate)
        elif opcode==0xB1:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._lda(arg, immediate)
        elif opcode==0xA2:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._ldx(arg, immediate)
        elif opcode==0xA6:
            cycles = 3
            arg = data[0]
            cycles += self._ldx(arg, immediate)
        elif opcode==0xB6:
            cycles = 4
            arg = (data[0] + self.Y) & 0xFF
            cycles += self._ldx(arg, immediate)
        elif opcode==0xAE:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._ldx(arg, immediate)
        elif opcode==0xBE:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._ldx(arg, immediate)
        elif opcode==0xA0:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._ldy(arg, immediate)
        elif opcode==0xA4:
            cycles = 3
            arg = data[0]
            cycles += self._ldy(arg, immediate)
        elif opcode==0xB4:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._ldy(arg, immediate)
        elif opcode==0xAC:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._ldy(arg, immediate)
        elif opcode==0xBC:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._ldy(arg, immediate)
        elif opcode==0x4A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._lsr(arg, immediate)
        elif opcode==0x46:
            cycles = 5
            arg = data[0]
            cycles += self._lsr(arg, immediate)
        elif opcode==0x56:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._lsr(arg, immediate)
        elif opcode==0x4E:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._lsr(arg, immediate)
        elif opcode==0x5E:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._lsr(arg, immediate)
        elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:
            cycles = 2
            arg = ARG_NONE
            cycles += self._nop(arg, immediate)
        elif opcode==0x09:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._ora(arg, immediate)
        elif opcode==0x05:
            cycles = 3
            arg = data[0]
            cycles += self._ora(arg, immediate)
        elif opcode==0x15:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._ora(arg, immediate)
        elif opcode==0x0D:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._ora(arg, immediate)
        elif opcode==0x1D:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._ora(arg, immediate)
        elif opcode==0x19:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._ora(arg, immediate)
        elif opcode==0x01:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._ora(arg, immediate)
        elif opcode==0x11:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._ora(arg, immediate)
        elif opcode==0x48:
            cycles = 3
            arg = ARG_NONE
            cycles += self._pha(arg, immediate)
        elif opcode==0x08:
            cycles = 3
            arg = ARG_NONE
            cycles += self._php(arg, immediate)
        elif opcode==0x68:
            cycles = 4
            arg = ARG_NONE
            cycles += self._pla(arg, immediate)
        elif opcode==0x28:
            cycles = 4
            arg = ARG_NONE
            cycles += self._plp(arg, immediate)
        elif opcode==0x2A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._rol(arg, immediate)
        elif opcode==0x26:
            cycles = 5
            arg = data[0]
            cycles += self._rol(arg, immediate)
        elif opcode==0x36:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._rol(arg, immediate)
        elif opcode==0x2E:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._rol(arg, immediate)
        elif opcode==0x3E:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._rol(arg, immediate)
        elif opcode==0x6A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._ror(arg, immediate)
        elif opcode==0x66:
            cycles = 5
            arg = data[0]
            cycles += self._ror(arg, immediate)
        elif opcode==0x76:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._ror(arg, immediate)
        elif opcode==0x6E:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._ror(arg, immediate)
        elif opcode==0x7E:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._ror(arg, immediate)
        elif opcode==0x40:
            cycles = 6
            arg = ARG_NONE
            cycles += self._rti(arg, immediate)
        elif opcode==0x60:
            cycles = 6
            arg = ARG_NONE
            cycles += self._rts(arg, immediate)
        elif opcode==0xE9 or opcode==0xEB:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._sbc(arg, immediate)
        elif opcode==0xE5:
            cycles = 3
            arg = data[0]
            cycles += self._sbc(arg, immediate)
        elif opcode==0xF5:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._sbc(arg, immediate)
        elif opcode==0xED:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._sbc(arg, immediate)
        elif opcode==0xFD:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._sbc(arg, immediate)
        elif opcode==0xF9:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._sbc(arg, immediate)
        elif opcode==0xE1:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._sbc(arg, immediate)
        elif opcode==0xF1:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._sbc(arg, immediate)
        elif opcode==0x38:
            cycles = 2
            arg = ARG_NONE
            cycles += self._sec(arg, immediate)
        elif opcode==0xF8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._sed(arg, immediate)
        elif opcode==0x78:
            cycles = 2
            arg = ARG_NONE
            cycles += self._sei(arg, immediate)
        elif opcode==0x85:
            cycles = 3
            arg = data[0]
            cycles += self._sta(arg, immediate)
        elif opcode==0x95:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._sta(arg, immediate)
        elif opcode==0x8D:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._sta(arg, immediate)
        elif opcode==0x9D:
            cycles = 5
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._sta(arg, immediate)
        elif opcode==0x99:
            cycles = 5
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._sta(arg, immediate)
        elif opcode==0x81:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._sta(arg, immediate)
        elif opcode==0x91:
            cycles = 6
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._sta(arg, immediate)
        elif opcode==0x86:
            cycles = 3
            arg = data[0]
            cycles += self._stx(arg, immediate)
        elif opcode==0x96:
            cycles = 4
            arg = (data[0] + self.Y) & 0xFF
            cycles += self._stx(arg, immediate)
        elif opcode==0x8E:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._stx(arg, immediate)
        elif opcode==0x84:
            cycles = 3
            arg = data[0]
            cycles += self._sty(arg, immediate)
        elif opcode==0x94:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._sty(arg, immediate)
        elif opcode==0x8C:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._sty(arg, immediate)
        elif opcode==0xAA:
            cycles = 2
            arg = ARG_NONE
            cycles += self._tax(arg, immediate)
        elif opcode==0xA8:
            cycles = 2
            arg = ARG_NONE
            cycles += self._tay(arg, immediate)
        elif opcode==0xBA:
            cycles = 2
            arg = ARG_NONE
            cycles += self._tsx(arg, immediate)
        elif opcode==0x8A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._txa(arg, immediate)
        elif opcode==0x9A:
            cycles = 2
            arg = ARG_NONE
            cycles += self._txs(arg, immediate)
        elif opcode==0x98:
            cycles = 2
            arg = ARG_NONE
            cycles += self._tya(arg, immediate)
        elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:
            cycles = 3
            arg = data[0]
            cycles += self._dop(arg, immediate)
        elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:
            cycles = 4
            arg = (data[0] + self.X) & 0xFF
            cycles += self._dop(arg, immediate)
        elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._dop(arg, immediate)
        elif opcode==0x0C:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._top(arg, immediate)
        elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:
            cycles = 4
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
            cycles += self._top(arg, immediate)
        elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:
            cycles = 1
            arg = ARG_NONE
            cycles += self._kil(arg, immediate)
        elif opcode==0xA7:
            cycles = 3
            arg = data[0]
            cycles += self._lax(arg, immediate)
        elif opcode==0xB7:
            cycles = 4
            arg = (data[0] + self.Y) & 0xFF
            cycles += self._lax(arg, immediate)
        elif opcode==0xAF:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._lax(arg, immediate)
        elif opcode==0xBF:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._lax(arg, immediate)
        elif opcode==0xA3:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._lax(arg, immediate)
        elif opcode==0xB3:
            cycles = 5
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
            cycles += self._lax(arg, immediate)
        elif opcode==0x87:
            cycles = 3
            arg = data[0]
            cycles += self._aax(arg, immediate)
        elif opcode==0x97:
            cycles = 4
            arg = (data[0] + self.Y) & 0xFF
            cycles += self._aax(arg, immediate)
        elif opcode==0x83:
            cycles = 6
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._aax(arg, immediate)
        elif opcode==0x8F:
            cycles = 4
            arg = self._from_le(data)
            cycles += self._aax(arg, immediate)
        elif opcode==0xC7:
            cycles = 5
            arg = data[0]
            cycles += self._dcp(arg, immediate)
        elif opcode==0xD7:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._dcp(arg, immediate)
        elif opcode==0xCF:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._dcp(arg, immediate)
        elif opcode==0xDF:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._dcp(arg, immediate)
        elif opcode==0xDB:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._dcp(arg, immediate)
        elif opcode==0xC3:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._dcp(arg, immediate)
        elif opcode==0xD3:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._dcp(arg, immediate)
        elif opcode==0xE7:
            cycles = 5
            arg = data[0]
            cycles += self._isc(arg, immediate)
        elif opcode==0xF7:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._isc(arg, immediate)
        elif opcode==0xEF:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._isc(arg, immediate)
        elif opcode==0xFF:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._isc(arg, immediate)
        elif opcode==0xFB:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._isc(arg, immediate)
        elif opcode==0xE3:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._isc(arg, immediate)
        elif opcode==0xF3:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._isc(arg, immediate)
        elif opcode==0x07:
            cycles = 5
            arg = data[0]
            cycles += self._slo(arg, immediate)
        elif opcode==0x17:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._slo(arg, immediate)
        elif opcode==0x0F:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._slo(arg, immediate)
        elif opcode==0x1F:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._slo(arg, immediate)
        elif opcode==0x1B:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._slo(arg, immediate)
        elif opcode==0x03:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._slo(arg, immediate)
        elif opcode==0x13:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._slo(arg, immediate)
        elif opcode==0x27:
            cycles = 5
            arg = data[0]
            cycles += self._rla(arg, immediate)
        elif opcode==0x37:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._rla(arg, immediate)
        elif opcode==0x2F:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._rla(arg, immediate)
        elif opcode==0x3F:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._rla(arg, immediate)
        elif opcode==0x3B:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._rla(arg, immediate)
        elif opcode==0x23:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._rla(arg, immediate)
        elif opcode==0x33:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._rla(arg, immediate)
        elif opcode==0x47:
            cycles = 5
            arg = data[0]
            cycles += self._sre(arg, immediate)
        elif opcode==0x57:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._sre(arg, immediate)
        elif opcode==0x4F:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._sre(arg, immediate)
        elif opcode==0x5F:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._sre(arg, immediate)
        elif opcode==0x5B:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._sre(arg, immediate)
        elif opcode==0x43:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._sre(arg, immediate)
        elif opcode==0x53:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._sre(arg, immediate)
        elif opcode==0x67:
            cycles = 5
            arg = data[0]
            cycles += self._rra(arg, immediate)
        elif opcode==0x77:
            cycles = 6
            arg = (data[0] + self.X) & 0xFF
            cycles += self._rra(arg, immediate)
        elif opcode==0x6F:
            cycles = 6
            arg = self._from_le(data)
            cycles += self._rra(arg, immediate)
        elif opcode==0x7F:
            cycles = 7
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._rra(arg, immediate)
        elif opcode==0x7B:
            cycles = 7
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._rra(arg, immediate)
        elif opcode==0x63:
            cycles = 8
            arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            cycles += self._rra(arg, immediate)
        elif opcode==0x73:
            cycles = 8
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._rra(arg, immediate)
        elif opcode==0x6B:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._arr(arg, immediate)
        elif opcode==0x4B:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._asr(arg, immediate)
        elif opcode==0xAB:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._atx(arg, immediate)
        elif opcode==0x0B or opcode==0x2B:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._aac(arg, immediate)
        elif opcode==0x9F:
            cycles = 5
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._axa(arg, immediate)
        elif opcode==0x93:
            cycles = 6
            m = self._read_word(data[0], wrap_at_page=True)
            arg = (m + self.Y) & 0xFFFF
            cycles += self._axa(arg, immediate)
        elif opcode==0xCB:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._axs(arg, immediate)
        elif opcode==0xBB:
            cycles = 4
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
            cycles += self._lar(arg, immediate)
        elif opcode==0x9E:
            cycles = 5
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._sxa(arg, immediate)
        elif opcode==0x9C:
            cycles = 5
            arg = (self._from_le(data) + self.X) & 0xFFFF
            cycles += self._sya(arg, immediate)
        elif opcode==0x8B:
            cycles = 2
            arg = data[0]
            immediate = True
            cycles += self._xaa(arg, immediate)
        elif opcode==0x9B:
            cycles = 5
            arg = (self._from_le(data) + self.Y) & 0xFFFF
            cycles += self._xas(arg, immediate)

        return cycles


import pyximport; pyximport.install()

from .cycore.carts import NESCart0, NESCart1, NESCart2, NESCart4
from nes.pycore.carts import NESCart0 as pyNESCart0
from .pycore.bitwise import upper_nibble, lower_nibble, bit_low, bit_high

class ROM:
    """
    Class for reading ROM data and generating cartridges from it.
    """
    # header byte 6
    MIRROR_BIT = 0
    PERSISTENT_BIT = 1
    TRAINER_BIT = 2
    MIRROR_IGNORE_BIT = 3

    # header byte 7
    NES2_FORMAT_MASK   = 0b00001100

    # mirror patterns
    # The mirror pattern specifies the underlying nametable at locations 0x2000, 0x2400, 0x2800 and 0x3200
    MIRROR_HORIZONTAL = [0, 0, 1, 1]
    MIRROR_VERTICAL = [0, 1, 0, 1]
    MIRROR_FOUR_SCREEN = [0, 1, 2, 3]

    def __init__(self, filename, verbose=True, py_compatibility_mode=False):
        self.py_compatibility_mode = py_compatibility_mode

        self.prg_rom_bytes = None
        self.chr_rom_bytes = None
        self.mirror_pattern = None
        self.mirror_ignore = None
        self.has_persistent = None
        self.has_trainer = None
        self.mapper_id = None
        self.submapper_id = None

        self.nes2 = None

        self.prg_ram_bytes = None
        self.prg_nvram_bytes = None
        self.chr_ram_bytes = None
        self.chr_nvram_bytes = None

        # the data itself
        self.trainer_data = None    # load at 0x7000  (not original data, used for compatibility code)
        self.prg_rom_data = None    # load at 0x8000
        self.chr_rom_data = None
        self.misc_rom_data = None

        # verbose?
        self.verbose = verbose

        if filename is not None:
            self.load(filename)

    def load(self, filename, ):
        """
        Load a ROM in the standard .nes file format
        """
        with open(filename, "rb") as f:
            nesheader = f.read(16)
            self.decode_header(nesheader)
            if self.has_trainer:
                self.trainer_data = f.read(512)
            self.prg_rom_data = f.read(self.prg_rom_bytes)
            self.chr_rom_data = f.read(self.chr_rom_bytes)
            self.misc_rom_data = f.read()  # likely to be empty in almost all cases
            if len(self.misc_rom_data) > 0:
                print("WARNING: MISC ROM DATA IS NOT EMPTY")

    def decode_header(self, nesheader):
        """
        Decode the standard .nes file format header.  Includes support for NES 2.0 format.
        """
        # header bytes 0-3 are fixed to ascii N E S <eof>
        #if not (nesheader[0] == 'N' and nesheader[1] == 'E' and nesheader[2] == 'S' and nesheader[3] == 0x1A):
            #raise ValueError("Invalid .nes file header")

        # header bytes 4 and 5
        self.prg_rom_bytes = nesheader[4] * 16384  # in 16kB banks
        self.chr_rom_bytes = nesheader[5] * 8192   # in 8kB banks

        # header byte 6
        self.mirror_pattern = self.MIRROR_HORIZONTAL if bit_low(nesheader[6], self.MIRROR_BIT) \
            else self.MIRROR_VERTICAL
        self.has_persistent = bit_high(nesheader[6], self.PERSISTENT_BIT)
        self.has_trainer = bit_high(nesheader[6], self.TRAINER_BIT)
        self.mirror_ignore = bit_high(nesheader[6], self.MIRROR_IGNORE_BIT)
        if self.mirror_ignore:  # if this is set, cart provides a 4-page vram
            self.mirror_pattern = self.MIRROR_FOUR_SCREEN

        # header byte 7
        self.mapper_id = upper_nibble(nesheader[7]) * 16 + upper_nibble(nesheader[6])
        self.nes2 = (nesheader[7] & self.NES2_FORMAT_MASK) > 0

        if not self.nes2:
            # header byte 8 (apparently often unused)
            self.prg_ram_bytes = max(1, nesheader[8]) * 8192
            if self.verbose:
                print("iNES (v1) Header")
        else:
            # NES 2.0 format
            # https://wiki.nesdev.com/w/index.php/NES_2.0
            if self.verbose:
                print("NES 2.0 Header")
            # header byte 8
            self.mapper_id += lower_nibble(nesheader[8]) * 256
            self.submapper_id = upper_nibble(nesheader[8])

            # header byte 9
            prg_rom_msb = lower_nibble(nesheader[9])
            if prg_rom_msb == 0xF:
                raise NotImplementedError("NES2 decoding for PRG_ROM_SIZE MSB nibble = 0xF")
            chr_rom_msb = upper_nibble(nesheader[9])
            if chr_rom_msb == 0xF:
                raise NotImplementedError("NES2 decoding for CHR_ROM_SIZE MSB nibble = 0xF")

            # header byte 10
            self.prg_ram_bytes = 64 << lower_nibble(nesheader[10])
            self.prg_nvram_bytes = 64 << upper_nibble(nesheader[10])

            # header byte 11
            self.chr_ram_bytes = 64 << lower_nibble(nesheader[11])
            self.chr_nvram_bytes = 64 << upper_nibble(nesheader[11])

        if self.verbose:
            print("Mapper: {}".format(self.mapper_id))
            print("prg_ram_bytes: {}".format(self.prg_ram_bytes))
            print("chr_ram_bytes: {}".format(self.chr_ram_bytes))
            print("prg_rom_bytes: {}".format(self.prg_rom_bytes))
            print("chr_rom_bytes: {}".format(self.chr_rom_bytes))
            print("mirror pattern: {}".format(self.mirror_pattern))
            #print("chr_ram_bytes: {}".format(self.chr_ram_bytes))

    def get_cart(self, interrupt_listener):
        """
        Get the correct type of cartridge object from this ROM, ready to be plugged into the NES system
        """
        if self.py_compatibility_mode:
            if self.mapper_id==0:
                return pyNESCart0(prg_rom_data=self.prg_rom_data,
                                chr_rom_data=self.chr_rom_data,
                                nametable_mirror_pattern=self.mirror_pattern,
                                )
            else:
                print("Mapper {} not currently supported in py_compatibility_mode".format(self.mapper_id))

        if self.mapper_id == 0:
            return NESCart0(prg_rom_data=self.prg_rom_data,
                            chr_rom_data=self.chr_rom_data,
                            nametable_mirror_pattern=self.mirror_pattern,
                            )
        elif self.mapper_id == 1:
            if self.chr_ram_bytes and (self.chr_ram_bytes != len(self.chr_rom_data)):
                raise ValueError("CHR RAM requested, but have not allocated the correct amount.")
            return NESCart1(prg_rom_data=self.prg_rom_data,
                            chr_rom_data=self.chr_rom_data,
                            prg_ram_size_kb=self.prg_ram_bytes / 1024,
                            chr_mem_writeable=True if self.chr_ram_bytes else False,
                            nametable_mirror_pattern=self.mirror_pattern,
                           )
        elif self.mapper_id == 2:
            return NESCart2(prg_rom_data=self.prg_rom_data,
                            chr_rom_data=self.chr_rom_data,
                            nametable_mirror_pattern=self.mirror_pattern,
                           )
        elif self.mapper_id == 4:
            # MMC 3
            return NESCart4(prg_rom_data=self.prg_rom_data,
                            chr_rom_data=self.chr_rom_data,
                            prg_ram_size_kb=self.prg_ram_bytes / 1024,
                            chr_mem_writeable=True if self.chr_ram_bytes else False,
                            nametable_mirror_pattern=self.mirror_pattern,
                            mirror_pattern_fixed=self.mirror_ignore,
                            interrupt_listener=interrupt_listener
                           )
        else:
            print("Mapper {} not currently supported".format(self.mapper_id))



import logging
import time

import pyximport; pyximport.install()

from nes.pycore.memory import NESVRAM
from nes.pycore.bitwise import bit_high, bit_low, set_bit, clear_bit, set_high_byte, set_low_byte
#from nes import LOG_PPU

class NESPPU:
    """
    NES Picture Processing Unit (PPU), the 2C02

    References:
        [1] Overall reference:  https://wiki.nesdev.com/w/index.php/PPU_programmer_reference
        [2] Rendering timing: https://wiki.nesdev.com/w/index.php/PPU_rendering
        [3] OAM layout:  https://wiki.nesdev.com/w/index.php/PPU_OAM

        [4] Detailed operation: http://nesdev.com/2C02%20technical%20operation.TXT

        [5] Palette generator: https://bisqwit.iki.fi/utils/nespalette.php

        [6] Register behaviour: https://wiki.nesdev.com/w/index.php/PPU_registers
        [7] Scrolling: https://wiki.nesdev.com/w/index.php/PPU_scrolling
    """
    NUM_REGISTERS = 8
    OAM_SIZE_BYTES = 256

    # Register indices
    # (this is not just an enum, this is the offset of the register in the CPU memory map from 0x2000)
    PPU_CTRL = 0
    PPU_MASK = 1
    PPU_STATUS = 2
    OAM_ADDR = 3
    OAM_DATA = 4
    PPU_SCROLL = 5
    PPU_ADDR = 6
    PPU_DATA = 7

    # masks for the bits in ppu registers
    # ppu_status
    VBLANK_MASK =               0b10000000  # same for ppu_ctrl
    SPRITE0_HIT_MASK =          0b01000000
    SPRITE_OVERFLOW_MASK =      0b00100000

    # ppu_ctrl
    SPRITE_SIZE_MASK =          0b00100000
    BKG_PATTERN_TABLE_MASK =    0b00010000
    SPRITE_PATTERN_TABLE_MASK = 0b00001000
    VRAM_INCREMENT_MASK =       0b00000100
    NAMETABLE_MASK =            0b00000011

    # ppu_mask
    RENDERING_ENABLED_MASK =    0b00011000
    RENDER_SPRITES_MASK =       0b00010000
    RENDER_BACKGROUND_MASK =    0b00001000
    RENDER_LEFT8_SPRITES_MASK = 0b00000100
    RENDER_LEFT8_BKG_MASK =     0b00000010
    GREYSCALE_MASK =            0b00000001


    # bit numbers of some important bits in registers
    # ppu_status
    V_BLANK_BIT = 7             # same for ppu_ctrl

    # ppu mask
    RENDER_LEFT8_BKG_BIT = 1
    RENDER_LEFT8_SPRITES_BIT = 2

    # byte numbers in ppu scroll
    PPU_SCROLL_X = 0
    PPU_SCROLL_Y = 1

    # screen and sprite/tile sizes:
    PIXELS_PER_LINE = 341       # number of pixels per ppu scanline; only 256 of thes are visible
    SCREEN_HEIGHT_PX = 240      # visible screen height (number of visible rows)
    SCREEN_WIDTH_PX = 256       # visible screen width (number of visible pixels per row)
    TILE_HEIGHT_PX = int(8)          # height of a tile/standard sprite in pixels
    TILE_WIDTH_PX = int(8)           # width of tile/standard sprite in pixels
    SCREEN_TILE_ROWS = 30       # number of rows of background tiles in a single screen
    SCREEN_TILE_COLS = 32       # number of columns of tiles in a single screen
    PATTERN_BITS_PER_PIXEL = 2  # number of bits used to represent each pixel in the patterns

    # the total size of a tile in the pattern table in bytes (== 16)
    PATTERN_SIZE_BYTES = int(TILE_WIDTH_PX * TILE_HEIGHT_PX * PATTERN_BITS_PER_PIXEL / 8)

    # A NES rgb palette mapping from NES color values to RGB; others are possible.
    DEFAULT_NES_PALETTE = [
        ( 82,  82,  82), (  1,  26,  81), ( 15,  15, 101), ( 35,   6,  99),
        ( 54,   3,  75), ( 64,   4,  38), ( 63,   9,   4), ( 50,  19,   0),
        ( 31,  32,   0), ( 11,  42,   0), (  0,  47,   0), (  0,  46,  10),
        (  0,  38,  45), (  0,   0,   0), (  0,   0,   0), (  0,   0,   0),
        (160, 160, 160), ( 30,  74, 157), ( 56,  55, 188), ( 88,  40, 184),
        (117,  33, 148), (132,  35,  92), (130,  46,  36), (111,  63,   0),
        ( 81,  82,   0), ( 49,  99,   0), ( 26, 107,   5), ( 14, 105,  46),
        ( 16,  92, 104), (  0,   0,   0), (  0,   0,   0), (  0,   0,   0),
        (254, 255, 255), (105, 158, 252), (137, 135, 255), (174, 118, 255),
        (206, 109, 241), (224, 112, 178), (222, 124, 112), (200, 145,  62),
        (166, 167,  37), (129, 186,  40), ( 99, 196,  70), ( 84, 193, 125),
        ( 86, 179, 192), ( 60,  60,  60), (  0,   0,   0), (  0,   0,   0),
        (254, 255, 255), (190, 214, 253), (204, 204, 255), (221, 196, 255),
        (234, 192, 249), (242, 193, 223), (241, 199, 194), (232, 208, 170),
        (217, 218, 157), (201, 226, 158), (188, 230, 174), (180, 229, 199),
        (181, 223, 228), (169, 169, 169), (  0,   0,   0), (  0,   0,   0),
    ]

    def __init__(self, cart=None, screen=None, interrupt_listener=None):

        # Registers
        self.ppu_ctrl = 0
        self.ppu_mask = 0
        self.oam_addr = 0
        self._oam_addr_held = 0         # this holds the oam_addr value at a certain point in the frame, when it is fixed for the whole frame
        self.oam_data = 0
        self.ppu_scroll = bytearray(2)  # this contains x-scroll (byte 0) and y-scroll (byte 1) accumulated over two writes
        self.ppu_addr = 0               # the accumulated **16-bit** address
        self._ppu_byte_latch = 0        # latch to keep track of which byte is being written in ppu_scroll and ppu_addr; latch is shared

        # internal latches to deal with open bus and buffering behaviour
        self._ppu_data_buffer = 0       # data to hold buffered reads from VRAM (see read of ppu_data)
        self._io_latch = 0              # last write/valid read of the ppu registers, sometimes reflected in read status

        # internal latches used in background rendering
        self._palette = [self.DEFAULT_NES_PALETTE[0:3], self.DEFAULT_NES_PALETTE[3:6]]     # 2 x palette latches
        self._pattern_lo = 0   # 16 bit patterns register to hold 2 x 8 bit patterns
        self._pattern_hi = 0   # 16 bit patterns register to hold 2 x 8 bit patterns

        # internal memory and latches used in sprite rendering
        self._oam = bytearray(32)      # this is a secondary internal array of OAM used to store sprite that will be active on the next scanline
        self._sprite_pattern = [[None] * 8 for _ in range(8)]
        self._sprite_bkg_priority = [0] * 8
        self._active_sprites = []

        # some state used in rendering to tell us where on the screen we are drawing
        self.line = 0
        self.pixel = 0
        self.row = 0
        self.col = 0
        #self._t = 0
        #self._tN = 0
        #self._tX = 0
        #self._tY = 0

        # internal statuses
        self.in_vblank = False
        self.sprite_zero_hit = False
        self.sprite_overflow = False

        # status used by emulator
        self.cycles_since_reset = 0
        self.cycles_since_frame = 0  # number of cycles since the frame start
        self.frames_since_reset = 0  # need all three counters (not really, but easier) because frame lengths vary
        self.time_at_new_frame = None

        # memory
        self.vram = NESVRAM(cart=cart)
        self.oam = bytearray(self.OAM_SIZE_BYTES)

        # screen attached to PPU
        self.screen = screen

        # interrupt listener
        self.interrupt_listener = interrupt_listener

        # palette: use the default, but can be replaced using utils.load_palette
        self.rgb_palette = self.DEFAULT_NES_PALETTE
        self.transparent_color = self._get_non_palette_color()
        self._palette_cache = [[None] * 4, [None] * 4]

        # tell the screen what rgb value the ppu is using to represent transparency
        if self.screen:
            self.screen.transparent_color = self.transparent_color

    def write_oam(self, data):
        for i in range(self.OAM_SIZE_BYTES):
            self.oam[i] = data[i]

    def invalidate_palette_cache(self):
        self._palette_cache = [[None] * 4, [None] * 4]

    def _get_non_palette_color(self):
        """
        Find a non-palette color in order to represent transparent pixels for blitting
        """
        trans_c = (1, 1, 1)
        while True:
            found = False
            for c in self.rgb_palette:
                if trans_c == c:
                    found = True
                    break
            if not found:
                return trans_c
            else:
                # just explore the grays, there are only 64 colors in palette, so even all
                # greys cannot be represented
                trans_c = (trans_c[0] + 1, trans_c[1] + 1, trans_c[2] + 1)

    @property
    def ppu_status(self):
        """
        The ppu status register value (without io latch noise in lower bits)
        :return:
        """
        return self.VBLANK_MASK * self.in_vblank \
               + self.SPRITE0_HIT_MASK * self.sprite_zero_hit \
               + self.SPRITE_OVERFLOW_MASK * self.sprite_overflow

    def read_register(self, register):
        """
        Read the specified PPU register (and take the correct actions along with that)
        This is mostly (always?) triggered by the CPU reading memory mapped ram at 0x2000-0x3FFF
        "Reading a nominally wrtie-only register will return the latch's current value" [6].
        The "latch" here refers to the capacitance
        of the PPU lines, which leads to some degree of "memory" on the lines, which will hold the last
        value written to a port (including read only ones), or the last value read from a read-only port
        """
        if register == self.PPU_CTRL:
            # write only
            print("WARNING: reading i/o latch")
            return self._io_latch
        elif register == self.PPU_MASK:
            # write only
            print("WARNING: reading i/o latch")
            return self._io_latch
        elif register == self.PPU_STATUS:
            # clear ppu_scroll and ppu_addr latches
            self._ppu_byte_latch = 0   # this is a shared latch between scroll and addr
            #self._ppu_scroll_ix = 0   # ^^^^
            v = self.ppu_status + (0x00011111 & self._io_latch)
            self.in_vblank = False  # clear vblank in ppu_status
            self._io_latch = v
            return v
        elif register == self.OAM_ADDR:
            # write only
            print("WARNING: reading i/o latch")
            return self._io_latch
        elif register == self.OAM_DATA:
            # todo: does not properly implement the weird results of this read during rendering
            v = self.oam[self.oam_addr]
            self._io_latch = v
            return v
        elif register == self.PPU_SCROLL:
            # write only
            print("WARNING: reading i/o latch")
            return self._io_latch
        elif register == self.PPU_ADDR:
            # write only
            print("WARNING: reading i/o latch")
            return self._io_latch
        elif register == self.PPU_DATA:
            if self.ppu_addr < self.vram.PALETTE_START:
                v = self._ppu_data_buffer
                self._ppu_data_buffer = self.vram.read(self.ppu_addr)
            else:
                v = self.vram.read(self.ppu_addr)
                # palette reads will return the palette without buffering, but will put the mirrored NT byte in the read buffer.
                # i.e. reading $3F00 will give you the palette entry at $3F00 and will put the byte in VRAM[$2F00] in the read buffer
                # source: http://forums.nesdev.com/viewtopic.php?t=1721
                self._ppu_data_buffer = self.vram.read(self.ppu_addr - 0x1000)
            self._increment_vram_address()
            self._io_latch = self._ppu_data_buffer
            return v

    def write_register(self, register, value):
        """
        Write one of the PPU registers with byte value and do whatever else that entails
        """
        # need to store the last write because it affects the value read on ppu_status
        # "Writing any value to any PPU port, even to the nominally read-only PPUSTATUS, will fill this latch"  [6]
        self._io_latch = value & 0xFF

        value &= 0xFF  # can only write a byte here

        if register == self.PPU_CTRL:
            # write only
            # writes to ppu_ctrl are ignored at first
            if self.cycles_since_reset < 29658:
                return
            # can trigger an immediate NMI if we are in vblank and the (allow) vblank NMI trigger flag is flipped high
            trigger_nmi = self.in_vblank \
                          and (value & self.VBLANK_MASK) > 0 \
                          and (self.ppu_ctrl & self.VBLANK_MASK) == 0
            self.ppu_ctrl = value
            if trigger_nmi:
                self._trigger_nmi()
        elif register == self.PPU_MASK:
            # write only
            self.ppu_mask = value
        elif register == self.PPU_STATUS:
            # read only
            pass
        elif register == self.OAM_ADDR:
            # write only
            self.oam_addr = value
        elif register == self.OAM_DATA:
            # read/write
            self.oam[self.oam_addr] = value
            # increment the OAM address and wrap around if necessary (wraps to start of page since OAM addr specifies
            # sub-page address)
            self.oam_addr = (self.oam_addr + 1) & 0xFF
        elif register == self.PPU_SCROLL:
            # write only
            self.ppu_scroll[self._ppu_byte_latch] = value
            # flip which byte is pointed to on each write; reset on ppu status read.  Latch shared with ppu_addr.
            self._ppu_byte_latch = 1 - self._ppu_byte_latch
        elif register == self.PPU_ADDR:
            # write only
            # high byte first
            if self._ppu_byte_latch == 0:
                self.ppu_addr = (self.ppu_addr & 0x00FF) + (value << 8)
                # Writes here overwrite the current nametable bits in ppu_ctrl (or at least overwrite bits in an
                # internal latch that is equivalent to this); see [7].  Some games, e.g. SMB, rely on this behaviour
                self.ppu_ctrl = (self.ppu_ctrl & 0b11111100) + ((value & 0b00001100) >> 2)
                # a write here also has a very unusual effect on the coarse and fine y scroll [7]
                self.ppu_scroll[self.PPU_SCROLL_Y] = (self.ppu_scroll[self.PPU_SCROLL_Y] & 0b00111100) + \
                                                     ((value & 0b00000011) << 6) + ((value & 0b00110000) >> 4)
            else:
                self.ppu_addr = (self.ppu_addr & 0xFF00) + value
                # writes here have a weird effect on the x and y scroll values [7]
                # here we just directly change the values of the scroll registers since they are write only and are used
                # only for this (rather than accumulating in a different internal latch _t like shown in [7]).  I think
                # that this is okay.
                self.ppu_scroll[self.PPU_SCROLL_X] = (self.ppu_scroll[self.PPU_SCROLL_X] & 0b00000111) + ((value & 0b00011111) << 3)
                self.ppu_scroll[self.PPU_SCROLL_Y] = (self.ppu_scroll[self.PPU_SCROLL_Y] & 0b11000111) + ((value & 0b11100000) >> 2)
            # flip which byte is pointed to on each write; reset on ppu status read
            self._ppu_byte_latch = 1 - self._ppu_byte_latch
        elif register == self.PPU_DATA:
            # read/write
            self.vram.write(self.ppu_addr, value)
            self._increment_vram_address()
            # invalidate the palette cache if this is a write to the palette memory.  Could be more careful about what
            # is invalidated here so that potentially less recalculation is needed.
            if self.ppu_addr >= self.vram.PALETTE_START:
                self.invalidate_palette_cache()

    def _increment_vram_address(self):
        """
        Increment vram address after reads/writes by an amount specified by a value in ppu_ctrl
        """
        self.ppu_addr += 1 if (self.ppu_ctrl & self.VRAM_INCREMENT_MASK) == 0 else 32

    def _trigger_nmi(self):
        """
        Do whatever is necessary to trigger an NMI to the CPU; note that it is up to the caller to check whether NMIs
        should be generated by the PPU at this time (a flag in ppu_ctrl), and respecting this is critically important.
        """
        self.interrupt_listener.raise_nmi()

    def _prefetch_active_sprites(self, line):
        """
        Non cycle-correct detector for active sprites on the given line.  Returns a list of the indices of the start
        address of the sprite in the OAM
        """
        # scan through the sprites, starting at oam_start_addr, seeing if they are visible in the line given
        # (note that should be the next line); if so, add them to the list of active sprites, until that gets full.
        # if using 8x16 sprites (True), or 8x8 sprite (False)
        double_sprites = (self.ppu_ctrl & self.SPRITE_SIZE_MASK) > 0
        sprite_height = 16 if double_sprites else 8

        self._active_sprites = []
        sprite_line = []
        for n in range(64):
            addr = (self._oam_addr_held + n * 4) % self.OAM_SIZE_BYTES
            sprite_y = self.oam[addr]
            if sprite_y <= line < sprite_y + sprite_height:
                self._active_sprites.append(addr)
                sprite_line.append(line - sprite_y)
                if len(self._active_sprites) >= 9:
                    break
        if len(self._active_sprites) > 8:
            # todo: this implements the *correct* behaviour of sprite overflow, but not the buggy behaviour
            # (and even then it is not cycle correct, so could screw up games that rely on timing of this very exactly)
            self.sprite_overflow = True
            self._active_sprites = self._active_sprites[:8]
            sprite_line = sprite_line[:8]

        self._fill_sprite_latches(self._active_sprites, sprite_line, double_sprites)

    def _fill_sprite_latches(self, active_sprite_addrs, sprite_line, double_sprites):
        """
        Non cycle-correct way to pre-fetch the sprite lines for the next scanline
        """
        table_base = ((self.ppu_ctrl & self.SPRITE_PATTERN_TABLE_MASK) > 0) * 0x1000

        for i, address in enumerate(active_sprite_addrs):
            attribs = self.oam[(address + 2) & 0xFF]
            palette_ix = attribs & 0b00000011
            palette = self.decode_palette(palette_ix, is_sprite=True)
            flip_v = bit_high(attribs, bit=7)
            flip_h = bit_high(attribs, bit=6)
            self._sprite_bkg_priority[i] = bit_high(attribs, bit=5)

            if not double_sprites:
                tile_ix = self.oam[(address + 1) & 0xFF]
                line = sprite_line[i] if not flip_v else 7 - sprite_line[i]
            else:
                line = sprite_line[i] if not flip_v else 15 - sprite_line[i]
                tile_ix = self.oam[(address + 1) & 0xFF] & 0b11111110
                if line >= 8:
                    # in the lower tile
                    tile_ix += 1
                    line -= 8

            tile_base = table_base + tile_ix * self.PATTERN_SIZE_BYTES
            sprite_pattern_lo = self.vram.read(tile_base + line)
            sprite_pattern_hi = self.vram.read(tile_base + 8 + line)

            for x in range(8):
                c = bit_high(sprite_pattern_hi, x) * 2 + bit_high(sprite_pattern_lo, x)
                self._sprite_pattern[i][x if flip_h else 7 - x] = palette[c] if c else None

            #print(self.line, i, sprite_line[i], self._sprite_pattern[i], self.oam[address + 3], self._sprite_bkg_priority[i])

    def _overlay_sprites(self, bkg_pixel):
        """
        Cycle-correct (ish) sprite rendering for the pixel at y=line, pixel=pixel.  Includes sprite 0 collision detection.
        """
        c_out = bkg_pixel
        if (self.ppu_mask & self.RENDER_SPRITES_MASK) == 0 \
            or (self.pixel - 1 < 8 and bit_low(self.ppu_mask, self.RENDER_LEFT8_SPRITES_BIT)):
            return c_out

        sprite_c_out, top_sprite = None, None
        s0_visible = False
        for i in reversed(range(len(self._active_sprites))):
            # render in reverse to make overwriting easier
            sprite_addr = self._active_sprites[i]
            sprite_x = self.oam[sprite_addr + 3]
            if sprite_x <= self.pixel - 1 < sprite_x + 8:
                #print(self.line, i, sprite_x, sprite_addr, self._sprite_bkg_priority[i])
                pix = self.pixel - 1 - sprite_x
                # this sprite is visible now
                c = self._sprite_pattern[i][pix]
                if c:
                    top_sprite = i
                    sprite_c_out = c
                    if sprite_addr == 0:
                        s0_visible = True

        # sprite zero collision detection
        # Details: https://wiki.nesdev.com/w/index.php/PPU_OAM#Sprite_zero_hits
        if s0_visible and bkg_pixel != self.transparent_color:
            # todo: there are some more fine details here
            self.sprite_zero_hit = True

        # now decide whether to keep sprite or bkg pixel
        if sprite_c_out and (not self._sprite_bkg_priority[top_sprite] or bkg_pixel == self.transparent_color):
            c_out = sprite_c_out

        return c_out if c_out != self.transparent_color else self.decode_palette(0)[0]  # background color

    def run_cycles(self, num_cycles):
        # cycles correspond to screen pixels during the screen-drawing phase of the ppu
        # there are three ppu cycles per cpu cycles, at least on NTSC systems
        frame_ended = False
        for cyc in range(num_cycles):
            # current scanline of the frame we are on - this determines behaviour during the line
            if self.line <= 239 and (self.ppu_mask & self.RENDERING_ENABLED_MASK) > 0:
                # visible scanline
                if 0 < self.pixel <= 256:  # pixels 1 - 256
                    # render pixel - 1
                    if (self.pixel - 1) % 8 == 0 and self.pixel > 1:
                        # fill background data latches
                        # todo: this is not cycle-correct, since the read is done atomically at the eighth pixel rather than throughout the cycle.
                        self.shift_bkg_latches()  # move the data from the upper latches into the current ones
                        self.fill_bkg_latches(self.line, self.col + 1)   # get some more data for the upper latches

                    # render background from latches
                    bkg_pixel = self._get_bkg_pixel()
                    # overlay srpite from latches
                    final_pixel = self._overlay_sprites(bkg_pixel)
                    if final_pixel != self.transparent_color:
                        self.screen.write_at(x=self.pixel - 1, y=self.line, color=final_pixel)
                    #self.shift_bkg_latches()
                elif self.pixel == 257:   # pixels 257 - 320
                    # sprite data fetching: fetch data from OAM for sprites on the next scanline
                    # NOTE:  "evaluation applies to the next line's sprite rendering, ... and this is why
                    # there is a 1 line offset on a sprite's Y coordinate."
                    # source: https://wiki.nesdev.com/w/index.php/PPU_sprite_evaluation  (note 1)
                    # this is implemented by passing line + 1 - 1 == line to the prefetch function for next line
                    self._prefetch_active_sprites(self.line)
                elif 321 <= self.pixel <= 336:   # pixels 321 - 336
                    # fill background data latches with data for first two tiles of next scanline
                    if self.pixel % 8 == 1:  # will happen at 321 and 329
                        # fill latches
                        self.shift_bkg_latches()  # move the data from the upper latches into the current ones
                        self.fill_bkg_latches(self.line + 1, int((self.pixel - 321) / 8))  # get some more data for the upper latches
                    #self.shift_bkg_latches()
                else:  # pixels 337 - 340
                    # todo: unknown nametable fetches (used by MMC5)
                    pass

            if self.line == 0 and self.pixel == 65:
                # The OAM address is fixed after this point  [citation needed]
                self._oam_addr_held = self.oam_addr
            elif self.line == 240 and self.pixel == 0:
                # post-render scanline, ppu is idle
                pass
            elif self.line == 241 and self.pixel == 1:
                # set vblank flag
                self.in_vblank = True   # set the vblank flag in ppu_status register
                # trigger NMI (if NMI is enabled)
                if (self.ppu_ctrl & self.VBLANK_MASK) > 0:
                    self._trigger_nmi()
            elif 241 <= self.line <= 260:
                # during vblank, ppu does no memory accesses; most of the CPU accesses happens here
                pass
            elif self.line == 261:
                # pre-render scanline for next frame; at dot 1, reset vblank flag in ppu_status
                if self.pixel == 1:
                    self.in_vblank = False
                    self.sprite_zero_hit = False
                    self.sprite_overflow = False
                elif self.pixel == 257:
                    # load sprite data for next scanline
                    # self._prefetch_active_sprites(line=0)
                    # "Sprite evaluation does not happen on the pre-render scanline. Because evaluation applies to the
                    # next line's sprite rendering, no sprites will be rendered on the first scanline, and this is why
                    # there is a 1 line offset on a sprite's Y coordinate."
                    # source: https://wiki.nesdev.com/w/index.php/PPU_sprite_evaluation  (note 1)
                    self._active_sprites = []
                elif 321 <= self.pixel <= 336:
                    # load data for next scanline
                    if self.pixel % 8 == 1:  # will happen at 321 and 329
                        # fill latches
                        self.shift_bkg_latches()  # move the data from the upper latches into the current ones
                        self.fill_bkg_latches(line=0, col=int((self.pixel - 321) / 8))  # get some more data for the upper latches
                elif self.pixel == self.PIXELS_PER_LINE - 1 - self.frames_since_reset % 2:
                    # this is the last pixel in the frame, so trigger the end-of-frame
                    # (do it below all the counter updates below, though)
                    frame_ended=True

            self.cycles_since_reset += 1
            self.cycles_since_frame += 1
            self.pixel += 1
            if self.pixel > 1 and self.pixel % 8 == 1:
                self.col += 1
            if self.pixel >= self.PIXELS_PER_LINE:
                self.line += 1
                self.pixel = 0
                self.col = 0
                if self.line > 0 and self.line % 8 == 0:
                    self.row += 1

            if frame_ended:
                self._new_frame()

            #logging.log(LOG_PPU, self.log_line(), extra={"source": "PPU"})
        return frame_ended

    def fill_bkg_latches(self, line, col):
        """
        Fill the ppu's rendering latches with the next tile to be rendered
        :return:
        """
        # todo: optimization - a lot of this can be precalculated once and then invalidated if anything changes, since that might only happen once per frame (or never)
        # get the tile from the nametable
        row = int(line / 8)

        # x and y coords of the nametable
        nx0 = bit_high(self.ppu_ctrl, 0)
        ny0 = bit_high(self.ppu_ctrl, 1)

        total_row = (row + ny0 * self.SCREEN_TILE_ROWS + ((self.ppu_scroll[self.PPU_SCROLL_Y] & 0b11111000) >> 3)) % (2 * self.SCREEN_TILE_ROWS)
        total_col = (col + nx0 * self.SCREEN_TILE_COLS + ((self.ppu_scroll[self.PPU_SCROLL_X] & 0b11111000) >> 3)) % (2 * self.SCREEN_TILE_COLS)

        ny = int(total_row / self.SCREEN_TILE_ROWS)
        nx = int(total_col / self.SCREEN_TILE_COLS)

        tile_row = total_row - ny * self.SCREEN_TILE_ROWS
        tile_col = total_col - nx * self.SCREEN_TILE_COLS

        ntbl_base = self.vram.NAMETABLE_START + (ny * 2 + nx) * self.vram.NAMETABLE_LENGTH_BYTES
        tile_addr = ntbl_base + tile_row * self.SCREEN_TILE_COLS + tile_col

        tile_index = self.vram.read(tile_addr)

        tile_bank = (self.ppu_ctrl & self.BKG_PATTERN_TABLE_MASK) > 0
        table_base = tile_bank * 0x1000
        tile_base = table_base + tile_index * self.PATTERN_SIZE_BYTES

        attribute_byte = self.vram.read(ntbl_base
                                        + self.vram.ATTRIBUTE_TABLE_OFFSET
                                        + (int(tile_row / 4) * 8 + int(tile_col / 4))
                                        )

        shift = 4 * (int(tile_row / 2) % 2) + 2 * (int(tile_col / 2) % 2)
        mask = 0b00000011 << shift
        palette_id = (attribute_byte & mask) >> shift

        self._palette[0] = self._palette[1]
        self._palette[1] = self.decode_palette(palette_id, is_sprite=False)

        tile_line = line - 8 * row

        self._pattern_lo = set_low_byte(self._pattern_lo, self.vram.read(tile_base + tile_line))
        self._pattern_hi = set_low_byte(self._pattern_hi, self.vram.read(tile_base + tile_line + 8))

    def shift_bkg_latches(self):
        self._pattern_hi <<= 8
        self._pattern_lo <<= 8

    def _get_bkg_pixel(self):
        if (   self.ppu_mask & self.RENDER_BACKGROUND_MASK) == 0 \
            or (self.pixel - 1 < 8 and bit_low(self.ppu_mask, self.RENDER_LEFT8_BKG_BIT)):
            return self.transparent_color

        fine_x = self.ppu_scroll[self.PPU_SCROLL_X] & 0b00000111
        px = (self.pixel - 1) % 8 + fine_x
        mask = 1 << (15 - px)
        v = ((self._pattern_lo & mask) > 0) + ((self._pattern_hi & mask) > 0) * 2
        return self._palette[int(px / 8)][v] if v > 0 else self.transparent_color

    def log_line(self):
        log = "{:5d}, {:3d}, {:3d}   ".format(self.frames_since_reset, self.line, self.pixel)
        log += "C:{:02X} M:{:02X} S:{:02X} OA:{:02X} OD:{:02X} ".format(self.ppu_ctrl,
                                                                                  self.ppu_mask,
                                                                                  self.ppu_status,
                                                                                  self.oam_addr,
                                                                                  self.oam_data)

        log += "SC:{:02X},{:02X} PA:{:04X}".format(self.ppu_scroll[0],
                                                                   self.ppu_scroll[1],
                                                                   self.ppu_addr)

        return log

    def _new_frame(self):
        """
        Things to do at the start of a frame
        """
        self.frames_since_reset += 1
        self.cycles_since_frame = 0
        self.pixel = 0
        self.line = 0
        self.row = 0
        self.col = 0

        #logging.log(logging.INFO, "PPU frame {} starting".format(self.frames_since_reset), extra={"source": "PPU"})
        # todo: "Vertical scroll bits are reloaded if rendering is enabled" - don't know what this means

    def decode_palette(self, palette_id, is_sprite=False):
        """
        If is_sprite is true, then decodes palette from the sprite palettes, otherwise
        decodes from the background palette tables.
        """
        if self._palette_cache[is_sprite][palette_id]:
            return self._palette_cache[is_sprite][palette_id]

        # get the palette colours (these are in hue (chroma) / value (luma) format.)
        # palette_id is in range 0..3, and gives an offset into one of the four background palettes,
        # each of which consists of three colors, each of which is represented by a singe byte
        palette_address = self.vram.PALETTE_START + 16 * is_sprite + 4 * palette_id
        palette = []
        for i in range(4):
            palette.append(self.rgb_palette[self.vram.read(palette_address + i) & 0b00111111])
        self._palette_cache[is_sprite][palette_id] = palette
        return palette
